#!/usr/bin/env python
"""
_WorkUnit_t_

Unit tests for the WMBS WorkUnit class.
"""

from __future__ import absolute_import, division, print_function

import threading
import unittest

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.WMBS.File import File
from WMCore.WMBS.WorkUnit import WorkUnit
from WMCore.WMBS.Workflow import Workflow
from WMQuality.TestInit import TestInit

WF_NAME = 'Test'


class WMBSWorkUnitTest(unittest.TestCase):
    """
    _WorkUnit_t_

    Unit tests for the WMBS WorkUnit class.
    """

    def __init__(self, *args, **kwargs):
        super(WMBSWorkUnitTest, self).__init__(*args, **kwargs)
        self.testWorkflow = None
        self.testFile = None

    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        WMBS tables.
        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"], useDefault=False)

        myThread = threading.currentThread()
        self.daofactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger, dbinterface=myThread.dbi)

        self.createPrerequisites()

        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """
        self.testInit.clearDatabase()
        return

    def createPrerequisites(self):
        """
        Create a dummy workflow we can use later
        """

        self.testWorkflow = Workflow(spec="spec.xml", owner="Simon", name="wf001", task=WF_NAME, wfType="ReReco")
        self.testWorkflow.create()

        self.testFile = File(lfn="/this/is/a/test/file0", size=1024, events=10)
        self.testFile.addRun(Run(1, *[44]))
        self.testFile.create()

        return

    def testCreateDeleteExists(self):
        """
        _testCreateDeleteExists_

        Test the create(), delete() and exists() methods of the WorkUnit class
        by creating and deleting a WorkUnit.  The exists() method will be
        called before and after creation and after deletion.
        """

        testWU = WorkUnit(wuid=1, taskID=self.testWorkflow.id, lastUnitCount=1, fileid=1, runLumi=Run(1, *[44]))
        sameWU = WorkUnit(taskID=self.testWorkflow.id, lastUnitCount=1, fileid=1, runLumi=Run(1, *[44]))

        self.assertFalse(testWU.exists(), "WorkUnit exists before it was created")
        self.assertFalse(sameWU.exists(), "WorkUnit exists before it was created")

        testWU.create()
        self.assertTrue(testWU.exists(), "WorkUnit does not exist after it was created")
        self.assertTrue(sameWU.exists(), "WorkUnit does not exist after it was created")

        testWU.delete()
        self.assertFalse(testWU.exists(), "WorkUnit exists after it has been deleted")
        self.assertFalse(sameWU.exists(), "WorkUnit exists after it has been deleted")

        return

    def testCreateTransaction(self):
        """
        _testCreateTransaction_

        Begin a transaction and then create a WorkUnit in the database.  Afterwards,
        rollback the transaction.  Use the WorkUnit class's exists() method to
        to verify that the file doesn't exist before it was created, exists
        after it was created and doesn't exist after the transaction was rolled
        back.
        """

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testWU = WorkUnit(wuid=1, taskID=self.testWorkflow.id, lastUnitCount=1, fileid=1, runLumi=Run(1, *[44]))

        self.assertFalse(testWU.exists(), "WorkUnit exists before it was created")

        testWU.create()
        self.assertTrue(testWU.exists(), "WorkUnit does not exist after it was created")

        myThread.transaction.rollback()
        self.assertFalse(testWU.exists(), "WorkUnit exists after rollback")

        return

    def testDeleteTransaction(self):
        """
        _testDeleteTransaction_

        Create a WorkUnit and commit it to the database.  Start a new transaction
        and delete the WorkUnit.  Rollback the transaction after the WorkUnit has been
        deleted.  Use the WorkUnit class's exists() method to verify that the file
        does not exist after it has been deleted but does exist after the
        transaction is rolled back.
        """

        testWU = WorkUnit(wuid=1, taskID=self.testWorkflow.id, lastUnitCount=1, fileid=1, runLumi=Run(1, *[44]))

        self.assertFalse(testWU.exists(), "WorkUnit exists before it was created")

        testWU.create()
        self.assertTrue(testWU.exists(), "WorkUnit does not exist after it was created")

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testWU.delete()
        self.assertFalse(testWU.exists(), "WorkUnit exists after it was deleted")

        myThread.transaction.rollback()
        self.assertTrue(testWU.exists(), "WorkUnit does not exist transaction was rolled back")

        return

    def testGetInfo(self):
        """
        _testGetInfo_

        Test the getInfo() method of the WorkUnit class to make sure that it
        returns the correct id. Everything else is tested by the tests in
        DataStructs_t which test the underlying getInfo()
        """

        testWU = WorkUnit(wuid=1, taskID=self.testWorkflow.id, lastUnitCount=1, fileid=1, runLumi=Run(1, *[44]))
        testWU.create()
        info = testWU.getInfo()
        self.assertEqual(info[0], testWU['id'], "Work unit returned wrong ID")

        return

    def testLoad(self):
        """
        _testLoad_

        Test the loading of WorkUnit data using the ID of a WorkUnit
        """

        testRetries = 10
        testSubmitTime = 10 * 365 * 24 * 3600
        testStatus = 4
        testFirstEvent = 100
        testLastEvent = 600
        testLastUnitCount = 4
        testRunLumi = Run(1, 44)

        intFields = ['id', 'taskid', 'retry_count', 'last_unit_count', 'last_submit_time', 'status',
                     'fileid', 'firstevent', 'lastevent']
        allFields = intFields + ['run_lumi']

        testWU = WorkUnit(wuid=1, taskID=self.testWorkflow.id, lastUnitCount=testLastUnitCount,
                          fileid=self.testFile['id'],
                          runLumi=testRunLumi, retryCount=testRetries, lastSubmitTime=testSubmitTime,
                          status=testStatus, firstEvent=testFirstEvent, lastEvent=testLastEvent)
        testWU.create()

        loadedWU = WorkUnit(wuid=testWU['id'])
        loadedWU.load()

        # Make sure load() gets the type set correctly
        for field in intFields:
            self.assertIsInstance(loadedWU[field], int, 'Field "%s" is not of type int' % field)

        # Make sure the values com back OK
        for field in allFields:
            self.assertEqual(testWU[field], loadedWU[field],
                             'Field "%s" is not returned correctly (got %s instead of %s)' %
                             (field, loadedWU[field], testWU[field]))

        # Now load things by taskid, fileid, run/lumi and run the same tests

        loadByFRL = WorkUnit(taskID=self.testWorkflow.id, fileid=self.testFile['id'], runLumi=testRunLumi)
        loadByFRL.load()

        # Make sure load() gets the type set correctly
        for field in intFields:
            self.assertIsInstance(loadByFRL[field], int, 'Field "%s" is not of type int' % field)

        # Make sure the values com back OK
        for field in allFields:
            self.assertEqual(testWU[field], loadByFRL[field],
                             'Field "%s" is not returned correctly (got %s instead of %s)' %
                             (field, loadByFRL[field], testWU[field]))
        testWU.delete()
        return

        # def testLoadData(self):
        #     """
        #     _testLoadData_
        #
        #     Test the loading of all data from a file, including run/lumi
        #     associations, location information and parentage information.
        #     """
        #     testFileParentA = File(lfn="/this/is/a/parent/lfnA", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentA.addRun(Run(1, *[45]))
        #     testFileParentB = File(lfn="/this/is/a/parent/lfnB", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentB.addRun(Run(1, *[45]))
        #     testFileParentA.create()
        #     testFileParentB.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #     testFileA.setLocation(pnn="T1_US_FNAL_Disk", immediateSave=False)
        #     testFileA.setLocation(pnn="T2_CH_CERN", immediateSave=False)
        #     testFileA.addParent("/this/is/a/parent/lfnA")
        #     testFileA.addParent("/this/is/a/parent/lfnB")
        #     testFileA.updateLocations()
        #
        #     testFileB = File(lfn=testFileA["lfn"])
        #     testFileB.loadData(parentage=1)
        #     testFileC = File(id=testFileA["id"])
        #     testFileC.loadData(parentage=1)
        #
        #     assert testFileA == testFileB, \
        #         "ERROR: File load by LFN didn't work"
        #
        #     assert testFileA == testFileC, \
        #         "ERROR: File load by ID didn't work"
        #
        #     testFileA.delete()
        #     testFileParentA.delete()
        #     testFileParentB.delete()
        #     return
        #
        # def testAddChild(self):
        #     """
        #     _testAddChild_
        #
        #     Add a child to some parent files and make sure that all the parentage
        #     information is loaded/stored correctly from the database.
        #     """
        #     testFileParentA = File(lfn="/this/is/a/parent/lfnA", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentA.addRun(Run(1, *[45]))
        #     testFileParentB = File(lfn="/this/is/a/parent/lfnB", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentB.addRun(Run(1, *[45]))
        #     testFileParentA.create()
        #     testFileParentB.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testFileParentA.addChild("/this/is/a/lfn")
        #     testFileParentB.addChild("/this/is/a/lfn")
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA, testFileParentB]
        #     for parentFile in testFileB["parents"]:
        #         assert parentFile in goldenFiles, \
        #             "ERROR: Unknown parent file"
        #         goldenFiles.remove(parentFile)
        #
        #     assert len(goldenFiles) == 0, \
        #         "ERROR: Some parents are missing"
        #     return
        #
        # def testAddChildTransaction(self):
        #     """
        #     _testAddChildTransaction_
        #
        #     Add a child to some parent files and make sure that all the parentage
        #     information is loaded/stored correctly from the database.  Rollback the
        #     addition of one of the childs and then verify that it does in fact only
        #     have one parent.
        #     """
        #     testFileParentA = File(lfn="/this/is/a/parent/lfnA", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentA.addRun(Run(1, *[45]))
        #     testFileParentB = File(lfn="/this/is/a/parent/lfnB", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentB.addRun(Run(1, *[45]))
        #     testFileParentA.create()
        #     testFileParentB.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testFileParentA.addChild("/this/is/a/lfn")
        #
        #     myThread = threading.currentThread()
        #     myThread.transaction.begin()
        #
        #     testFileParentB.addChild("/this/is/a/lfn")
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA, testFileParentB]
        #     for parentFile in testFileB["parents"]:
        #         assert parentFile in goldenFiles, \
        #             "ERROR: Unknown parent file"
        #         goldenFiles.remove(parentFile)
        #
        #     assert len(goldenFiles) == 0, \
        #         "ERROR: Some parents are missing"
        #
        #     myThread.transaction.rollback()
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA]
        #     for parentFile in testFileB["parents"]:
        #         assert parentFile in goldenFiles, \
        #             "ERROR: Unknown parent file"
        #         goldenFiles.remove(parentFile)
        #
        #     assert len(goldenFiles) == 0, \
        #         "ERROR: Some parents are missing"
        #
        #     return
        #
        # def testCreateWithLocation(self):
        #     """
        #     _testCreateWithLocation_
        #
        #     Create a file and add a couple locations.  Load the file from the
        #     database to make sure that the locations were set correctly.
        #     """
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1},
        #                      locations=set(["T1_US_FNAL_Disk", "T2_CH_CERN"]))
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData()
        #
        #     goldenLocations = ["T1_US_FNAL_Disk", "T2_CH_CERN"]
        #
        #     for location in testFileB["locations"]:
        #         assert location in goldenLocations, \
        #             "ERROR: Unknown file location"
        #         goldenLocations.remove(location)
        #
        #     assert len(goldenLocations) == 0, \
        #         "ERROR: Some locations are missing"
        #     return
        #
        # def testSetLocation(self):
        #     """
        #     _testSetLocation_
        #
        #     Create a file and add a couple locations.  Load the file from the
        #     database to make sure that the locations were set correctly.
        #     """
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testFileA.setLocation(["T1_US_FNAL_Disk", "T2_CH_CERN"])
        #     testFileA.setLocation(["bunkT1_US_FNAL_Disk", "bunkT2_CH_CERN"],
        #                           immediateSave=False)
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData()
        #
        #     goldenLocations = ["T1_US_FNAL_Disk", "T2_CH_CERN"]
        #
        #     for location in testFileB["locations"]:
        #         assert location in goldenLocations, \
        #             "ERROR: Unknown file location"
        #         goldenLocations.remove(location)
        #
        #     assert len(goldenLocations) == 0, \
        #         "ERROR: Some locations are missing"
        #     return
        #
        # def testSetLocationTransaction(self):
        #     """
        #     _testSetLocationTransaction_
        #
        #     Create a file at specific locations and commit everything to the
        #     database.  Reload the file from the database and verify that the
        #     locations are correct.  Rollback the database transaction and once
        #     again reload the file.  Verify that the original locations are back.
        #     """
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #     testFileA.setLocation(["T1_US_FNAL_Disk"])
        #
        #     myThread = threading.currentThread()
        #     myThread.transaction.begin()
        #
        #     testFileA.setLocation(["T2_CH_CERN"])
        #     testFileA.setLocation(["bunkT1_US_FNAL_Disk", "bunkT2_CH_CERN"],
        #                           immediateSave=False)
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData()
        #
        #     goldenLocations = ["T1_US_FNAL_Disk", "T2_CH_CERN"]
        #
        #     for location in testFileB["locations"]:
        #         assert location in goldenLocations, \
        #             "ERROR: Unknown file location"
        #         goldenLocations.remove(location)
        #
        #     assert len(goldenLocations) == 0, \
        #         "ERROR: Some locations are missing"
        #
        #     myThread.transaction.rollback()
        #     testFileB.loadData()
        #
        #     goldenLocations = ["T1_US_FNAL_Disk"]
        #
        #     for location in testFileB["locations"]:
        #         assert location in goldenLocations, \
        #             "ERROR: Unknown file location"
        #         goldenLocations.remove(location)
        #
        #     assert len(goldenLocations) == 0, \
        #         "ERROR: Some locations are missing"
        #     return
        #
        # def testLocationsConstructor(self):
        #     """
        #     _testLocationsConstructor_
        #
        #     Test to make sure that locations passed into the File() constructor
        #     are loaded from and save to the database correctly.  Also test to make
        #     sure that the class behaves well when the location is passed in as a
        #     single string instead of a set.
        #     """
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations=set(["T1_US_FNAL_Disk"]))
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testFileB = File(lfn="/this/is/a/lfn2", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileB.addRun(Run(1, *[45]))
        #     testFileB.create()
        #
        #     testFileC = File(id=testFileA["id"])
        #     testFileC.loadData()
        #
        #     goldenLocations = ["T1_US_FNAL_Disk"]
        #     for location in testFileC["locations"]:
        #         assert location in goldenLocations, \
        #             "ERROR: Unknown file location"
        #         goldenLocations.remove(location)
        #
        #     assert len(goldenLocations) == 0, \
        #         "ERROR: Some locations are missing"
        #
        #     testFileC = File(id=testFileB["id"])
        #     testFileC.loadData()
        #
        #     goldenLocations = ["T1_US_FNAL_Disk"]
        #     for location in testFileC["locations"]:
        #         assert location in goldenLocations, \
        #             "ERROR: Unknown file location"
        #         goldenLocations.remove(location)
        #
        #     assert len(goldenLocations) == 0, \
        #         "ERROR: Some locations are missing"
        #     return
        #
        # def testSetLocationOrder(self):
        #     """
        #     _testSetLocationOrder_
        #
        #     This tests that you can specify a location before creating the file,
        #     instead of having to do it afterwards.
        #     """
        #     myThread = threading.currentThread()
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10)
        #     testFileA.setLocation("T2_CH_CERN")
        #     testFileA.create()
        #
        #     testFileB = File(lfn=testFileA["lfn"])
        #     testFileB.load()
        #
        #     daoFactory = DAOFactory(package="WMCore.WMBS", logger=logging, dbinterface=myThread.dbi)
        #
        #     locationFac = daoFactory(classname="Files.GetLocation")
        #     location = locationFac.execute(testFileB['lfn']).pop()
        #
        #     self.assertEqual(location, 'T2_CH_CERN')
        #
        #     return
        #
        # def testAddRunSet(self):
        #     """
        #     _testAddRunSet_
        #
        #     Test the ability to add run and lumi information to a file.
        #     """
        #     testFile = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                     checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFile.create()
        #     runSet = set()
        #     runSet.add(Run(1, *[45]))
        #     runSet.add(Run(2, *[67, 68]))
        #     testFile.addRunSet(runSet)
        #
        #     assert (runSet - testFile["runs"]) == set(), \
        #         "Error: addRunSet is not updating set correctly"
        #
        #     return
        #
        # def testGetAncestorLFNs(self):
        #     """
        #     _testGenAncestorLFNs_
        #
        #     Create a series of files that have several generations of parentage
        #     information.  Verify that the parentage information is reported
        #     correctly.
        #     """
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileA.create()
        #
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileB.create()
        #
        #     testFileC = File(lfn="/this/is/a/lfnC", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileC.create()
        #
        #     testFileD = File(lfn="/this/is/a/lfnD", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileD.create()
        #
        #     testFileE = File(lfn="/this/is/a/lfnE", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileE.create()
        #
        #     testFileE = File(lfn="/this/is/a/lfnF", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T1_US_FNAL_Disk")
        #     testFileE.create()
        #
        #     testFileA.addParent(lfn="/this/is/a/lfnB")
        #     testFileA.addParent(lfn="/this/is/a/lfnC")
        #     testFileB.addParent(lfn="/this/is/a/lfnD")
        #     testFileC.addParent(lfn="/this/is/a/lfnD")
        #     testFileD.addParent(lfn="/this/is/a/lfnE")
        #     testFileD.addParent(lfn="/this/is/a/lfnF")
        #
        #     level1 = ["/this/is/a/lfnB", "/this/is/a/lfnC"]
        #     level2 = ["/this/is/a/lfnD"]
        #     level3 = ["/this/is/a/lfnE", "/this/is/a/lfnF"]
        #     level4 = level5 = []
        #
        #     decs2 = ["/this/is/a/lfnA"]
        #
        #     assert testFileA.getAncestors(level=1, type='lfn') == level1, \
        #         "ERROR: level 1 test failed"
        #     assert testFileA.getAncestors(level=2, type='lfn') == level2, \
        #         "ERROR: level 2 test failed"
        #     assert testFileA.getAncestors(level=3, type='lfn') == level3, \
        #         "ERROR: level 3 test failed"
        #     assert testFileA.getAncestors(level=4, type='lfn') == level4, \
        #         "ERROR: level 4 test failed"
        #     assert testFileA.getAncestors(level=5, type='lfn') == level5, \
        #         "ERROR: level 5 test failed"
        #
        #     assert testFileD.getDescendants(level=1, type='lfn') == level1, \
        #         "ERROR: level 1 desc test failed"
        #     assert testFileD.getDescendants(level=2, type='lfn') == decs2, \
        #         "ERROR: level 2 desc test failed"
        #     assert testFileD.getDescendants(level=3, type='lfn') == level4, \
        #         "ERROR: level 3 desc test failed"
        #
        #     return
        #
        # def testGetLocationBulk(self):
        #     """
        #     _testGetLocationBulk_
        #
        #     Checks to see whether the code that we have will enable us to get the locations
        #     of all files at once
        #     """
        #     myThread = threading.currentThread()
        #
        #     daoFactory = DAOFactory(package="WMCore.WMBS", logger=logging, dbinterface=myThread.dbi)
        #     locationAction = daoFactory(classname="Locations.New")
        #     locationAction.execute(siteName="site3", pnn="T2_CH_CERN2")
        #     locationAction.execute(siteName="site4", pnn="T2_CH_CERN3")
        #     locationAction.execute(siteName="site5", pnn="T2_CH_CERN4")
        #     locationAction.execute(siteName="site6", pnn="T2_CH_CERN5")
        #     locationAction.execute(siteName="site7", pnn="T2_CH_CERN6")
        #
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T2_CH_CERN")
        #     testFileA.create()
        #
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T2_CH_CERN2")
        #     testFileB.create()
        #
        #     testFileC = File(lfn="/this/is/a/lfnC", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T2_CH_CERN3")
        #     testFileC.create()
        #
        #     testFileD = File(lfn="/this/is/a/lfnD", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T2_CH_CERN4")
        #     testFileD.create()
        #
        #     testFileE = File(lfn="/this/is/a/lfnE", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T2_CH_CERN5")
        #     testFileE.create()
        #
        #     testFileF = File(lfn="/this/is/a/lfnF", size=1024, events=10,
        #                      checksums={'cksum': 1}, locations="T2_CH_CERN6")
        #     testFileF.create()
        #
        #     files = [testFileA, testFileB, testFileC, testFileD, testFileE, testFileF]
        #
        #     locationFac = daoFactory(classname="Files.GetLocationBulk")
        #     location = locationFac.execute(files=files)
        #
        #     for f in files:
        #         self.assertEqual(location[f.exists()], list(f['locations']))
        #
        #     return
        #
        # def testBulkParentage(self):
        #     """
        #     _testBulkParentage_
        #
        #     Verify that the bulk parentage dao correctly sets file parentage.
        #     """
        #     testFileChildA = File(lfn="/this/is/a/child/lfnA", size=1024,
        #                           events=20, checksums={'cksum': 1})
        #     testFileChildB = File(lfn="/this/is/a/child/lfnB", size=1024,
        #                           events=20, checksums={'cksum': 1})
        #     testFileChildA.create()
        #     testFileChildB.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.create()
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileB.create()
        #     testFileC = File(lfn="/this/is/a/lfnC", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileC.create()
        #
        #     parentage = [{"child": testFileChildA["id"], "parent": testFileA["id"]},
        #                  {"child": testFileChildA["id"], "parent": testFileB["id"]},
        #                  {"child": testFileChildA["id"], "parent": testFileC["id"]},
        #                  {"child": testFileChildB["id"], "parent": testFileA["id"]},
        #                  {"child": testFileChildB["id"], "parent": testFileB["id"]}]
        #
        #     myThread = threading.currentThread()
        #     daofactory = DAOFactory(package="WMCore.WMBS",
        #                             logger=myThread.logger,
        #                             dbinterface=myThread.dbi)
        #     bulkParentageAction = daofactory(classname="Files.AddBulkParentage")
        #     bulkParentageAction.execute(parentage)
        #
        #     testFileD = File(id=testFileChildA["id"])
        #     testFileD.loadData(parentage=1)
        #     testFileE = File(id=testFileChildB["id"])
        #     testFileE.loadData(parentage=1)
        #
        #     goldenFiles = [testFileA, testFileB, testFileC]
        #     for parentFile in testFileD["parents"]:
        #         assert parentFile in goldenFiles, \
        #             "ERROR: Unknown parent file"
        #         goldenFiles.remove(parentFile)
        #
        #     assert len(goldenFiles) == 0, \
        #         "ERROR: Some parents are missing"
        #
        #     goldenFiles = [testFileA, testFileB]
        #     for parentFile in testFileE["parents"]:
        #         assert parentFile in goldenFiles, \
        #             "ERROR: Unknown parent file"
        #         goldenFiles.remove(parentFile)
        #
        #     assert len(goldenFiles) == 0, \
        #         "ERROR: Some parents are missing"
        #     return
        #
        # def testDataStructsFile(self):
        #     """
        #     _testDataStructsFile_
        #
        #     Tests our ability to create a WMBS file from a DataStructs File and vice versa
        #     """
        #
        #     myThread = threading.currentThread()
        #
        #     testLFN = "lfn1"
        #     testSize = 1024
        #     testEvents = 100
        #     testCksum = {"cksum": '1'}
        #     testParents = set(["lfn2"])
        #     testRun = Run(1, *[45])
        #     testSE = "T2_CH_CERN"
        #
        #     parentFile = File(lfn="lfn2")
        #     parentFile.create()
        #
        #     testFile = File()
        #
        #     inputFile = WMFile(lfn=testLFN, size=testSize, events=testEvents, checksums=testCksum, parents=testParents)
        #     inputFile.addRun(testRun)
        #     inputFile.setLocation(pnn=testSE)
        #
        #     testFile.loadFromDataStructsFile(file=inputFile)
        #     testFile.create()
        #     testFile.save()
        #
        #     loadFile = File(lfn="lfn1")
        #     loadFile.loadData(parentage=1)
        #
        #     self.assertEqual(loadFile['size'], testSize)
        #     self.assertEqual(loadFile['events'], testEvents)
        #     self.assertEqual(loadFile['checksums'], testCksum)
        #     self.assertEqual(loadFile['locations'], set([testSE]))
        #     # self.assertEqual(loadFile['parents'].pop()['lfn'], 'lfn2')
        #
        #     wmFile = loadFile.returnDataStructsFile()
        #     self.assertEqual(wmFile == inputFile, True)
        #
        #     return
        #
        # def testParentageByJob(self):
        #     """
        #     _testParentageByJob_
        #
        #     Tests the DAO that assigns parentage by Job
        #     """
        #
        #     testWorkflow = Workflow(spec='hello', owner="mnorman",
        #                             name="wf001", task="basicWorkload/Production")
        #     testWorkflow.create()
        #     testFileset = Fileset(name="TestFileset")
        #     testFileset.create()
        #     testSubscription = Subscription(fileset=testFileset, workflow=testWorkflow, type="Processing",
        #                                     split_algo="FileBased")
        #     testSubscription.create()
        #     testJobGroup = JobGroup(subscription=testSubscription)
        #     testJobGroup.create()
        #
        #     testFileParentA = File(lfn="/this/is/a/parent/lfnA", size=1024,
        #                            events=20, checksums={'cksum': 1},
        #                            locations=set(['T2_CH_CERN', 'T1_US_FNAL_Disk']))
        #     testFileParentA.addRun(Run(1, *[45]))
        #     testFileParentB = File(lfn="/this/is/a/parent/lfnB", size=1024,
        #                            events=20, checksums={'cksum': 1},
        #                            locations=set(['T2_CH_CERN', 'T1_US_FNAL_Disk']))
        #     testFileParentB.addRun(Run(1, *[45]))
        #     testFileParentA.create()
        #     testFileParentB.create()
        #     testFileset.addFile(testFileParentA)
        #     testFileset.addFile(testFileParentB)
        #     testFileset.commit()
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testJobA = Job()
        #     testJobA["outcome"] = 'success'
        #     testJobA.create(group=testJobGroup)
        #     testJobA.addFile(testFileParentA)
        #     testJobA.addFile(testFileParentB)
        #     testJobA.associateFiles()
        #     testSubscription.acquireFiles()
        #     testJobA.completeInputFiles()
        #
        #     parentAction = self.daofactory(classname="Files.SetParentageByJob")
        #     parentAction.execute(binds={'jobid': testJobA.exists(), 'child': testFileA['lfn']})
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA, testFileParentB]
        #     for parentFile in testFileB["parents"]:
        #         self.assertEqual(parentFile in goldenFiles, True,
        #                          "ERROR: Unknown parent file")
        #         goldenFiles.remove(parentFile)
        #
        #     self.assertEqual(len(goldenFiles), 0,
        #                      "ERROR: Some parents are missing")
        #
        #     testFileC = File(lfn="/this/is/c/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileC.addRun(Run(1, *[46]))
        #     testFileC.create()
        #
        #     testJobC = Job()
        #     testJobC["outcome"] = 'failure'
        #     testJobC.create(group=testJobGroup)
        #     testJobC.addFile(testFileParentA)
        #     testJobC.associateFiles()
        #     testSubscription.acquireFiles()
        #     testJobC.failInputFiles()
        #
        #     parentAction.execute(binds={'jobid': testJobC.exists(), 'child': testFileC['lfn']})
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA, testFileParentB]
        #     for parentFile in testFileB["parents"]:
        #         self.assertEqual(parentFile in goldenFiles, True,
        #                          "ERROR: Unknown parent file")
        #
        #     testFileC_1 = File(id=testFileC["id"])
        #     testFileC_1.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA]
        #     for parentFile in testFileC_1["parents"]:
        #         self.assertEqual(parentFile in goldenFiles, True,
        #                          "ERROR: Unknown parent file")
        #
        # def testParentageByMergeJob(self):
        #     """
        #     _testParentageByJob_
        #
        #     Tests the DAO that assigns parentage by Job
        #     """
        #
        #     testWorkflow = Workflow(spec='hello', owner="mnorman",
        #                             name="wf001", task="basicWorkload/Production")
        #     testWorkflow.create()
        #     testFileset = Fileset(name="TestFileset")
        #     testFileset.create()
        #     testSubscription = Subscription(fileset=testFileset, workflow=testWorkflow, type="Processing",
        #                                     split_algo="FileBased")
        #     testSubscription.create()
        #     testJobGroup = JobGroup(subscription=testSubscription)
        #     testJobGroup.create()
        #
        #     testFileParentA = File(lfn="/this/is/a/parent/lfnA", size=1024,
        #                            events=20, checksums={'cksum': 1},
        #                            locations=set(['T2_CH_CERN', 'T1_US_FNAL_Disk']))
        #     testFileParentA.addRun(Run(1, *[45]))
        #     testFileParentB = File(lfn="/this/is/a/parent/lfnB", size=1024,
        #                            events=20, checksums={'cksum': 1},
        #                            locations=set(['T2_CH_CERN', 'T1_US_FNAL_Disk']))
        #     testFileParentB.addRun(Run(1, *[45]))
        #     testFileParentA.create()
        #     testFileParentB.create()
        #     testFileset.addFile(testFileParentA)
        #     testFileset.addFile(testFileParentB)
        #     testFileset.commit()
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #
        #     testJobA = Job()
        #     testJobA["outcome"] = 'success'
        #     testJobA.create(group=testJobGroup)
        #     testJobA.addFile(testFileParentA)
        #     testJobA.addFile(testFileParentB)
        #     testJobA.associateFiles()
        #     testSubscription.acquireFiles()
        #     testJobA.completeInputFiles()
        #
        #     parentAction = self.daofactory(classname="Files.SetParentageByMergeJob")
        #     parentAction.execute(binds={'jobid': testJobA.exists(), 'child': testFileA['lfn']})
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA, testFileParentB]
        #     for parentFile in testFileB["parents"]:
        #         self.assertEqual(parentFile in goldenFiles, True,
        #                          "ERROR: Unknown parent file")
        #         goldenFiles.remove(parentFile)
        #
        #     self.assertEqual(len(goldenFiles), 0,
        #                      "ERROR: Some parents are missing")
        #
        #     testFileC = File(lfn="/this/is/c/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileC.addRun(Run(1, *[46]))
        #     testFileC.create()
        #
        #     testJobC = Job()
        #     testJobC["outcome"] = 'failure'
        #     testJobC.create(group=testJobGroup)
        #     testJobC.addFile(testFileParentA)
        #     testJobC.associateFiles()
        #     testSubscription.acquireFiles()
        #     testJobC.failInputFiles()
        #
        #     parentAction.execute(binds={'jobid': testJobC.exists(), 'child': testFileC['lfn']})
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA]
        #     for parentFile in testFileB["parents"]:
        #         self.assertEqual(parentFile in goldenFiles, False,
        #                          "ERROR: Unknown parent file")
        #
        #     testFileC_1 = File(id=testFileC["id"])
        #     testFileC_1.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA]
        #     for parentFile in testFileC_1["parents"]:
        #         self.assertEqual(parentFile in goldenFiles, False,
        #                          "ERROR: Unknown parent file")
        #
        # def testAddChecksumsByLFN(self):
        #     """
        #     _testAddChecksumsByLFN_
        #
        #     Tests for adding checksums by DAO by LFN
        #     """
        #
        #     testWorkflow = Workflow(spec='hello', owner="mnorman",
        #                             name="wf001", task="basicWorkload/Production")
        #     testWorkflow.create()
        #     testFileset = Fileset(name="TestFileset")
        #     testFileset.create()
        #     testSubscription = Subscription(fileset=testFileset, workflow=testWorkflow, type="Processing",
        #                                     split_algo="FileBased")
        #     testSubscription.create()
        #     testJobGroup = JobGroup(subscription=testSubscription)
        #     testJobGroup.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        #     testFileB.addRun(Run(1, *[45]))
        #     testFileB.create()
        #
        #     testJobA = Job()
        #     testJobA.create(group=testJobGroup)
        #     testJobA.associateFiles()
        #
        #     parentAction = self.daofactory(classname="Files.AddChecksumByLFN")
        #     binds = [{'lfn': testFileA['lfn'], 'cktype': 'cksum', 'cksum': 101},
        #              {'lfn': testFileA['lfn'], 'cktype': 'adler32', 'cksum': 201},
        #              {'lfn': testFileB['lfn'], 'cktype': 'cksum', 'cksum': 101}]
        #     parentAction.execute(bulkList=binds)
        #
        #     testFileC = File(id=testFileA["id"])
        #     testFileC.loadData()
        #     testFileD = File(id=testFileB["id"])
        #     testFileD.loadData()
        #
        #     self.assertEqual(testFileC['checksums'], {'adler32': '201', 'cksum': '101'})
        #     self.assertEqual(testFileD['checksums'], {'cksum': '101'})
        #
        #     return
        #
        # def testSetLocationByLFN(self):
        #     """
        #     _testSetLocationByLFN_
        #
        #     Create a file and add a couple locations.  Load the file from the
        #     database to make sure that the locations were set correctly.
        #     """
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10,
        #                      checksums={'cksum': 1})
        #     testFileB.addRun(Run(1, *[45]))
        #     testFileB.create()
        #
        #     parentAction = self.daofactory(classname="Files.SetLocationByLFN")
        #     binds = [{'lfn': "/this/is/a/lfnA", 'location': 'T1_US_FNAL_Disk'},
        #              {'lfn': "/this/is/a/lfnB", 'location': 'T1_US_FNAL_Disk'}]
        #     parentAction.execute(lfn=binds)
        #
        #     testFileC = File(id=testFileA["id"])
        #     testFileC.loadData()
        #     testFileD = File(id=testFileB["id"])
        #     testFileD.loadData()
        #
        #     self.assertEqual(testFileC['locations'], set(['T1_US_FNAL_Disk']))
        #     self.assertEqual(testFileD['locations'], set(['T1_US_FNAL_Disk']))
        #
        #     return
        #
        # def testCreateWithParent(self):
        #
        #     """
        #     Test passing parnents arguments in file creation.
        #     check if parent file does not exist, it create the file and set the parentage
        #     """
        #
        #     # create parent file before it got added to child file.
        #     testFileParentA = File(lfn="/this/is/a/parent/lfnA", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentA.addRun(Run(1, *[45]))
        #     testFileParentA.create()
        #
        #     # don't create create parent file before it got added to child file.
        #     testFileParentB = File(lfn="/this/is/a/parent/lfnB", size=1024,
        #                            events=20, checksums={'cksum': 1})
        #     testFileParentB.addRun(Run(1, *[45]))
        #
        #     testFileA = File(lfn="/this/is/a/lfn", size=1024, events=10,
        #                      checksums={'cksum': 1},
        #                      parents=[testFileParentA, testFileParentB])
        #     testFileA.addRun(Run(1, *[45]))
        #
        #     testFileA.create()
        #
        #     testFileB = File(id=testFileA["id"])
        #     testFileB.loadData(parentage=1)
        #
        #     goldenFiles = [testFileParentA, testFileParentB]
        #     for parentFile in testFileB["parents"]:
        #         assert parentFile in goldenFiles, \
        #             "ERROR: Unknown parent file"
        #         goldenFiles.remove(parentFile)
        #
        #     assert len(goldenFiles) == 0, \
        #         "ERROR: Some parents are missing"
        #
        # def testAddToFileset(self):
        #     """
        #     _AddToFileset_
        #
        #     Test to see if we can add to a fileset using the DAO
        #     """
        #     testFileset = Fileset(name="inputFileset")
        #     testFileset.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        #     testFileB.addRun(Run(1, *[45]))
        #     testFileB.create()
        #
        #     addToFileset = self.daofactory(classname="Files.AddToFileset")
        #     addToFileset.execute(file=[testFileA['lfn'], testFileB['lfn']],
        #                          fileset=testFileset.id)
        #
        #     testFileset2 = Fileset(name="inputFileset")
        #     testFileset2.loadData()
        #
        #     self.assertEqual(len(testFileset2.files), 2)
        #     for file in testFileset2.files:
        #         self.assertTrue(file in [testFileA, testFileB])
        #
        #     # Check that adding twice doesn't crash
        #     addToFileset.execute(file=[testFileA['lfn'], testFileB['lfn']],
        #                          fileset=testFileset.id)
        #
        # def testAddDupsToFileset(self):
        #     """
        #     _AddToDupsFileset_
        #
        #     Verify the the dups version of the AddToFileset DAO will not add files
        #     to a fileset if they're already associated to another fileset with the
        #     same workflow.
        #     """
        #     testWorkflowA = Workflow(spec='hello', owner="mnorman",
        #                              name="wf001", task="basicWorkload/Production")
        #     testWorkflowA.create()
        #     testWorkflowB = Workflow(spec='hello', owner="mnorman",
        #                              name="wf001", task="basicWorkload/Production2")
        #     testWorkflowB.create()
        #
        #     testFilesetA = Fileset(name="inputFilesetA")
        #     testFilesetA.create()
        #     testFilesetB = Fileset(name="inputFilesetB")
        #     testFilesetB.create()
        #
        #     testSubscriptionA = Subscription(workflow=testWorkflowA, fileset=testFilesetA)
        #     testSubscriptionA.create()
        #     testSubscriptionB = Subscription(workflow=testWorkflowB, fileset=testFilesetB)
        #     testSubscriptionB.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileA.create()
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        #     testFileB.addRun(Run(1, *[45]))
        #     testFileB.create()
        #
        #     addToFileset = self.daofactory(classname="Files.AddDupsToFileset")
        #     addToFileset.execute(file=[testFileA['lfn'], testFileB['lfn']],
        #                          fileset=testFilesetA.id, workflow="wf001")
        #
        #     testFileset2 = Fileset(name="inputFilesetA")
        #     testFileset2.loadData()
        #
        #     self.assertEqual(len(testFileset2.files), 2)
        #     for file in testFileset2.files:
        #         self.assertTrue(file in [testFileA, testFileB])
        #
        #     # Check that adding twice doesn't crash
        #     addToFileset.execute(file=[testFileA['lfn'], testFileB['lfn']],
        #                          fileset=testFilesetA.id, workflow="wf001")
        #
        #     # Files should not get added to fileset B because fileset A is associated
        #     # with wf001.
        #     addToFileset.execute(file=[testFileA['lfn'], testFileB['lfn']],
        #                          fileset=testFilesetB.id, workflow="wf001")
        #
        #     testFileset2 = Fileset(name="inputFilesetB")
        #     testFileset2.loadData()
        #
        #     self.assertEqual(len(testFileset2.files), 0)
        #     return
        #
        # def testAddDupsToFilesetBulk(self):
        #     """
        #     _AddToDupsFilesetBulk_
        #
        #     Same as testAddDupsToFileset() but faster
        #     """
        #     testWorkflowA = Workflow(spec='hello', owner="mnorman",
        #                              name="wf001", task="basicWorkload/Production")
        #     testWorkflowA.create()
        #     testWorkflowB = Workflow(spec='hello', owner="mnorman",
        #                              name="wf001", task="basicWorkload/Production2")
        #     testWorkflowB.create()
        #
        #     testFilesetA = Fileset(name="inputFilesetA")
        #     testFilesetA.create()
        #     testFilesetB = Fileset(name="inputFilesetB")
        #     testFilesetB.create()
        #
        #     testSubscriptionA = Subscription(workflow=testWorkflowA, fileset=testFilesetA)
        #     testSubscriptionA.create()
        #     testSubscriptionB = Subscription(workflow=testWorkflowB, fileset=testFilesetB)
        #     testSubscriptionB.create()
        #
        #     testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10, locations=['SiteA'])
        #     testFileA.addRun(Run(1, *[45]))
        #     testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10, locations=['SiteB'])
        #     testFileB.addRun(Run(1, *[45]))
        #
        #     addFilesToWMBSInBulk(testFilesetA.id, "wf001",
        #                          [testFileA, testFileB],
        #                          conn=testFileA.getDBConn(),
        #                          transaction=testFileA.existingTransaction())
        #
        #     testFileset2 = Fileset(name="inputFilesetA")
        #     testFileset2.loadData()
        #
        #     self.assertEqual(len(testFileset2.files), 2)
        #     for file in testFileset2.files:
        #         self.assertTrue(file in [testFileA, testFileB])
        #
        #     # Check that adding twice doesn't crash
        #     addFilesToWMBSInBulk(testFilesetA.id, "wf001",
        #                          [testFileA, testFileB],
        #                          conn=testFileA.getDBConn(),
        #                          transaction=testFileA.existingTransaction())
        #
        #     # Files should not get added to fileset B because fileset A is associated
        #     # with wf001.
        #     addFilesToWMBSInBulk(testFilesetB.id, "wf001",
        #                          [testFileA, testFileB],
        #                          conn=testFileA.getDBConn(),
        #                          transaction=testFileA.existingTransaction())
        #
        #     testFileset2 = Fileset(name="inputFilesetB")
        #     testFileset2.loadData()
        #
        #     self.assertEqual(len(testFileset2.files), 0)
        #     return
        #
        # def test_SetLocationsForWorkQueue(self):
        #     """
        #     _SetLocationsForWorkQueue_
        #
        #     Test the code that sets locations for the WorkQueue
        #     This is more complicated then it seems.
        #     """
        #
        #     action = self.daofactory(classname="Files.SetLocationForWorkQueue")
        #
        #     testFile = File(lfn="myLFN", size=1024,
        #                     events=10, checksums={'cksum': 1111})
        #     testFile.create()
        #
        #     tFile1 = File(lfn="myLFN")
        #     tFile1.loadData()
        #     locations = tFile1.getLocations()
        #
        #     self.assertEqual(locations, [])
        #
        #     binds = [{'lfn': 'myLFN', 'location': 'T2_CH_CERN'}]
        #     action.execute(lfns=['myLFN'], locations=binds)
        #
        #     tFile1.loadData()
        #     locations = tFile1.getLocations()
        #     self.assertEqual(locations, ['T2_CH_CERN'])
        #
        #     binds = [{'lfn': 'myLFN', 'location': 'T1_US_FNAL_Disk'}]
        #     action.execute(lfns=['myLFN'], locations=binds)
        #
        #     tFile1.loadData()
        #     locations = tFile1.getLocations()
        #     self.assertEqual(locations, ['T1_US_FNAL_Disk'])
        #
        #     return


if __name__ == "__main__":
    unittest.main()



from WMCore.REST.Error import RESTError
from WMCore.REST.Error import NotAcceptable
from WMCore.REST.Error import UnsupportedMethod
from WMCore.REST.Error import MethodWithoutQueryString
from WMCore.REST.Error import APIMethodMismatch
from WMCore.REST.Error import APINotSpecified
from WMCore.REST.Error import NoSuchInstance
from WMCore.REST.Error import DatabaseError
from WMCore.REST.Error import DatabaseUnavailable
from WMCore.REST.Error import DatabaseConnectionError
from WMCore.REST.Error import DatabaseExecutionError
from WMCore.REST.Error import MissingParameter
from WMCore.REST.Error import InvalidParameter
from WMCore.REST.Error import MissingObject
from WMCore.REST.Error import TooManyObjects
from WMCore.REST.Error import ObjectAlreadyExists
from WMCore.REST.Error import InvalidObject
from WMCore.REST.Error import ExecutionError
RESTError()
NotAcceptable()
UnsupportedMethod()
MethodWithoutQueryString()
APIMethodMismatch()
APINotSpecified()
NoSuchInstance()
DatabaseError()
DatabaseUnavailable()
DatabaseConnectionError()
DatabaseExecutionError()
MissingParameter()
InvalidParameter()
MissingObject()
TooManyObjects()
ObjectAlreadyExists()
InvalidObject()
ExecutionError()



#!/usr/bin/env python

from builtins import object
from future import standard_library
standard_library.install_aliases()

import unittest
import logging
import urllib.request
from WMCore.WebTools.Root import Root
from WMCore.Configuration import Configuration
from cherrypy import engine, tree
from cherrypy import config as cpconfig
from tempfile import NamedTemporaryFile

# DISABLING because this doesn't properly shut down the cherrypy
# server or clean up the state
class RootTest(object):

    def getBaseConfiguration(self):
        config = Configuration()
        config.component_('SecurityModule')
        config.SecurityModule.dangerously_insecure = True

        config.component_('Webtools')
        config.Webtools.application = 'UnitTests'
        config.Webtools.log_screen = False
        config.Webtools.environment = "development"
        config.Webtools.error_log_level = logging.WARNING
        config.Webtools.access_log_level = logging.DEBUG
        config.Webtools.host = "127.0.0.1"
        config.component_('UnitTests')
        config.UnitTests.admin = "Mr Unit Test"
        config.UnitTests.description = "Dummy server for unit tests"
        config.UnitTests.title = "Unit Tests"
        config.UnitTests.section_('views')
        active = config.UnitTests.views.section_('active')

        return config

    def testLongHandConfigurables(self):
        """
        Test that the following configuration variables work:

        engine      Controls the "application engine", including autoreload.
                    These can only be declared in the global config.
        hooks       Declares additional request-processing functions.
        log         Configures the logging for each application. These can only
                    be declared in the global or / config.
        request         Adds attributes to each Request.
        response        Adds attributes to each Response.
        server      Controls the default HTTP server via cherrypy.server. These
                    can only be declared in the global config.
        tools       Runs and configures additional request-processing packages.
        wsgi        Adds WSGI middleware to an Application's "pipeline". These
                    can only be declared in the app's root config ("/").
        checker         Controls the "checker", which looks for common errors in app
                    state (including config) when the engine starts. Global config only.

        (from http://docs.cherrypy.org/dev/intro/concepts/config.html)
        """
        config = self.getBaseConfiguration()

        server = Root(config)
        server.start(blocking=False)
        server.stop()

    def testFakeLongHandConfigurables(self):
        """
        Test that a made up long hand configurable is ignored
        """
        config = self.getBaseConfiguration()
        # The following should be ignored by the configure step
        config.Webtools.section_('foo')
        config.Webtools.foo.bar = 'baz'
        config.Webtools.section_('stuff')
        config.Webtools.stuff = 'things'

        server = Root(config)
        server.start(blocking=False)

        self.assertFalse('foo' in cpconfig, 'non-standard configurable passed to server')
        self.assertFalse('stuff' in cpconfig, 'non-standard configurable passed to server')

        server.stop()

    def testMissingRequiredConfigParams(self):
        """
        All applications should define:
        ['admin', 'description', 'title']
        """
        config = self.getBaseConfiguration()

        config.UnitTests.__delattr__('admin')
        server = Root(config)
        self.assertRaises(AssertionError, server.start, blocking=False)

        config.UnitTests.admin = "Mr Unit Test"
        config.UnitTests.__delattr__('description')
        server = Root(config)
        self.assertRaises(AssertionError, server.start, blocking=False)

        config.UnitTests.description = "Dummy server for unit tests"
        config.UnitTests.__delattr__('title')
        server = Root(config)
        self.assertRaises(AssertionError, server.start, blocking=False)

    def testLongHandProxyBase(self):
        """
        Check that changing the proxy base via tools.proxy.base
        does actually change the proxy base
        """
        test_proxy_base = '/unit_test'

        config = self.getBaseConfiguration()
        config.Webtools.section_('tools')
        config.Webtools.tools.section_('proxy')
        config.Webtools.tools.proxy.base = test_proxy_base
        config.Webtools.tools.proxy.on = True
        server = Root(config)

        server.start(blocking=False)
        self.assertEqual(cpconfig['tools.proxy.base'], test_proxy_base)
        server.stop()

    def testShortHandProxyBase(self):
        """
        Check that changing the proxy_base via the short hand config variable
        does actually change the proxy base
        """
        test_proxy_base = '/unit_test'

        config = self.getBaseConfiguration()
        # Set the proxy base with a short hand cfg variable
        config.Webtools.proxy_base = test_proxy_base
        server = Root(config)

        server.start(blocking=False)
        self.assertEqual(cpconfig['tools.proxy.base'], test_proxy_base)
        server.stop()

    def testLongHandChangePort(self):
        """
        Change the port the server runs on long hand
        """
        test_port = 8010

        config = self.getBaseConfiguration()
        # Set the port to a non-standard one
        config.Webtools.section_('server')
        config.Webtools.server.socket_port = test_port

        server = Root(config)
        server.start(blocking=False)
        self.assertEqual(cpconfig['server.socket_port'], test_port)
        server.stop()

    def testShortHandChangePort(self):
        """
        Change the port the server runs on short hand
        """
        test_port = 8010

        config = self.getBaseConfiguration()
        # Set the port to a non-standard one
        config.Webtools.port = test_port

        server = Root(config)
        server.start(blocking=False)
        self.assertEqual(cpconfig['server.socket_port'], test_port)
        server.stop()

    def testShortHandPortOverride(self):
        """
        Change the port the server runs on long hand, then over ride
        it with the short hand equivalent
        """
        test_port = 8010

        config = self.getBaseConfiguration()
        # Set the port the long handed way
        config.Webtools.section_('server')
        config.Webtools.server.socket_port = test_port - 1
        # then override
        config.Webtools.port = test_port

        server = Root(config)
        server.start(blocking=False)
        self.assertEqual(cpconfig['server.socket_port'], test_port)
        server.stop()

    def testSecuritySetting(self):
        testRole = "TestRole"
        testGroup = "TestGroup"
        testSite = "TestSite"
        config = self.getBaseConfiguration()
        config.SecurityModule.dangerously_insecure = False
        # not real keyfile but for the test.
        # file will be deleted automaticall when garbage collected.
        tempFile = NamedTemporaryFile()
        config.SecurityModule.key_file = tempFile.name
        config.SecurityModule.section_("default")
        config.SecurityModule.default.role = testRole
        config.SecurityModule.default.group = testGroup
        config.SecurityModule.default.site = testSite
        config.Webtools.environment = "production"
        server = Root(config)
        server.start(blocking=False)
        self.assertEqual(cpconfig['tools.secmodv2.on'], True)
        self.assertEqual(cpconfig['tools.secmodv2.role'], testRole)
        self.assertEqual(cpconfig['tools.secmodv2.group'], testGroup)
        self.assertEqual(cpconfig['tools.secmodv2.site'], testSite)
        server.stop()

    def testInstanceInUrl(self):
        config = self.getBaseConfiguration()
        config.SecurityModule.dangerously_insecure = True
        server = Root(config)
        # Add our test page
        config.UnitTests.instances = ['foo', 'bar', 'baz/zoink']
        active = config.UnitTests.views.section_('active')
        active.section_('test')
        active.test.object = 'WMCore_t.WebTools_t.InstanceTestPage'
        active.test.section_('database')
        db_instances = active.test.database.section_('instances')
        foo = db_instances.section_('foo')
        bar = db_instances.section_('bar')
        baz = db_instances.section_('baz/zoink')
        foo.connectUrl = 'sqlite:///foo'
        bar.connectUrl = 'sqlite:///bar'
        baz.connectUrl = 'sqlite:///baz/zoink'
        active.test.section_('security')
        security_instances = active.test.security.section_('instances')
        sec_foo = security_instances.section_('foo')
        sec_bar = security_instances.section_('bar')
        sec_baz = security_instances.section_('baz/zoink')
        sec_foo.sec_params = 'test_foo'
        sec_bar.sec_params = 'test_bar'
        sec_baz.sec_params = 'test_baz'

        server.start(blocking=False)

        for instance in config.UnitTests.instances:
            url = 'http://127.0.0.1:%s/unittests/%s/test' % (cpconfig['server.socket_port'], instance)
            html = urllib.request.urlopen(url).read()
            self.assertEqual(html, instance)
            db_url = '%s/database' % url
            html = urllib.request.urlopen(db_url).read()
            self.assertEqual(html, db_instances.section_(instance).connectUrl)
            sec_url = '%s/security' % url
            html = urllib.request.urlopen(sec_url).read()
            self.assertEqual(html, security_instances.section_(instance).sec_params)
        server.stop()

    def testUsingFilterTool(self):
        """
        Use the filter tool to prevent unexpected accesses from
        unsupported methods
        TODO
        """
        pass

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_WorkQueueTestCase_

Unit tests for the WMBS File class.
"""

import logging
import os

from WMCore.Database.CMSCouch import CouchMonitor
from WMCore.Database.CMSCouch import CouchServer
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit


class WorkQueueTestCase(EmulatedUnitTestCase):

    def setSchema(self):
        "this can be override if the schema setting is different"
        self.schema = ["WMCore.WMBS","WMComponent.DBS3Buffer","WMCore.BossAir"]
        self.couchApps = ["WorkQueue"]

    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        WMBS tables.  Also add some dummy locations.
        """
        super(WorkQueueTestCase, self).setUp()
        self.queueDB = 'workqueue_t'
        self.queueInboxDB = 'workqueue_t_inbox'
        self.globalQDB = 'workqueue_t_global'
        self.globalQInboxDB = 'workqueue_t_global_inbox'
        self.localQDB = 'workqueue_t_local'
        self.localQInboxDB = 'workqueue_t_local_inbox'
        self.localQDB2 = 'workqueue_t_local2'
        self.localQInboxDB2 = 'workqueue_t_local2_inbox'
        self.configCacheDB = 'workqueue_t_config_cache'
        self.logDBName = 'logdb_t'
        self.requestDBName = 'workqueue_t_reqmgr_workload_cache'

        self.setSchema()
        self.testInit = TestInit('WorkQueueTest')
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase=True)
        self.addCleanup(self.testInit.clearDatabase)
        self.addCleanup(logging.debug, 'Cleanup called clearDatabase()')
        self.testInit.setSchema(customModules = self.schema,
                                useDefault = False)
        self.testInit.setupCouch(self.queueDB, *self.couchApps)
        self.testInit.setupCouch(self.queueInboxDB, *self.couchApps)
        self.testInit.setupCouch(self.globalQDB, *self.couchApps)
        self.testInit.setupCouch(self.globalQInboxDB , *self.couchApps)
        self.testInit.setupCouch(self.localQDB, *self.couchApps)
        self.testInit.setupCouch(self.localQInboxDB, *self.couchApps)
        self.testInit.setupCouch(self.localQDB2, *self.couchApps)
        self.testInit.setupCouch(self.localQInboxDB2, *self.couchApps)
        self.testInit.setupCouch(self.configCacheDB, 'ConfigCache')
        self.testInit.setupCouch(self.logDBName, 'LogDB')
        self.testInit.setupCouch(self.requestDBName, 'ReqMgr')

        self.couchURL = os.environ.get("COUCHURL")
        couchServer = CouchServer(self.couchURL)
        self.configCacheDBInstance = couchServer.connectDatabase(self.configCacheDB)

        self.localCouchMonitor = CouchMonitor(self.couchURL)
        self.localCouchMonitor.deleteReplicatorDocs()
        self.addCleanup(self.localCouchMonitor.deleteReplicatorDocs)
        self.addCleanup(logging.debug, 'Cleanup called deleteReplicatorDocs()')
        self.addCleanup(self.testInit.tearDownCouch)
        self.addCleanup(logging.debug, 'Cleanup called tearDownCouch()')

        self.workDir = self.testInit.generateWorkDir()
        self.addCleanup(self.testInit.delWorkDir)
        self.addCleanup(logging.debug, 'Cleanup called delWorkDir()')

        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """

        super(WorkQueueTestCase, self).tearDown()  # Left here in case it's needed by any of the sub-classes
        return



"""
_DBS3API_

Mock DBS3 api for testing the DBS3 uploader

Created on Mar 13, 2013

@author: dballest
"""

from builtins import object
import json
import os

from random import random

class DbsApi(object):
    """
    _DbsApi_

    Replacement DbsApi implementing the required
    functions for the DBS3 upload poller.
    """
    def __init__(self, url, **kwds):
        """
        _init_

        Store the url where this dbs is "located"

        :param url: url of the DBS server
        :param kwds: optional DBSApi keyword arguments which are irrelevant
        for mocking but should be presented nevertheless for compatibility with
        actual DBSApi class
        """
        self.dbsPath = url

    def insertBulkBlock(self, blockDump):
        """
        _insertBulkBlock_

        Insert a block into fake DBS, only insert the block specific information
        and no file information in the file given in the dbsPath.
        There is 20% chance of a proxy error!
        """

        currentInfo = []
        if os.path.getsize(self.dbsPath):
            with open(self.dbsPath, 'r') as inFileHandle:
                currentInfo = json.load(inFileHandle)
        with open(self.dbsPath, 'w') as outFileHandle:
            for block in currentInfo:
                if block["block"]["block_name"] == blockDump["block"]["block_name"]:
                    raise Exception("Block %s already exists" % blockDump["block"]["block_name"])
            currentInfo.append(blockDump)
            json.dump(currentInfo, outFileHandle)

        # VK: once we introduced parseDBSException function we do not need to simulate Proxy Error
        # see https://github.com/dmwm/WMCore/pull/11176#issuecomment-1158758332
        # randomNumber = random()
        # if randomNumber < 0.2:
        #     raise Exception("Proxy Error, this is a mock proxy error.")

        return

    def listBlocks(self, block_name):
        """
        _listBlocks_

        Return the requested block information if it exists.
        """
        if os.path.getsize(self.dbsPath):
            with open(self.dbsPath, 'r') as inFileHandle:
                currentInfo = json.load(inFileHandle)
                inFileHandle.close()
                for block in currentInfo:
                    if block["block"]["block_name"] == block_name:
                        return [block["block"]]
        return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Script to produce a json file to mock Rucio data
"""

from __future__ import (division, print_function)

import json
import os
import sys

from future.utils import viewitems

from WMCore.Services.Rucio.Rucio import Rucio
from WMCore.WMBase import getTestBase


### Here goes a list of data that we want to fetch from Rucio and
### persist in our json file to mock those calls/data
CONTAINERS = [u"/MinimumBias/ComissioningHI-v1/RAW",
              u"/Cosmics/ComissioningHI-PromptReco-v1/RECO"]
BLOCKS = []

### The output file which will contain all the Rucio mock data
OUTPUT_FILE = "RucioMockData.json"

### some default constants (which will probably not change in a long time)
SCOPE = "cms"
RUCIO_URL = 'http://cms-rucio.cern.ch'
RUCIO_AUTH_URL = 'https://cms-rucio-auth.cern.ch'
RUCIO_ACCT = "wmcore_transferor"


def main():
    finalResult = {}
    rucio = Rucio(RUCIO_ACCT,
                  hostUrl=RUCIO_URL,
                  authUrl=RUCIO_AUTH_URL,
                  configDict={"auth_type": "x509"})
    realRucio = rucio.cli
    outFilename = os.path.join(getTestBase(), '..', 'data', 'Mock', OUTPUT_FILE)

    calls = []
    for container in CONTAINERS:
        print("Building call list for container: {}".format(container))
        calls.append(['getBlocksInContainer', {'container': container}])
        calls.append(['isContainer', {'didName': container}])
        calls.append(['getDID', {'didName': container, 'dynamic': False}])
        calls.append(['didExist', {'didName': container}])
        blocks = realRucio.list_content(scope=SCOPE, name=container)
        for block in blocks:
            if block['type'].upper() != 'DATASET':
                continue
            calls.append(['isContainer', {'didName': block['name']}])
            calls.append(['getDID', {'didName': block['name'], 'dynamic': False}])
            calls.append(['didExist', {'didName': block['name']}])
        for block in BLOCKS:
            calls.append(['isContainer', {'didName': block}])
            calls.append(['getDID', {'didName': block, 'dynamic': False}])
            calls.append(['didExist', {'didName': block}])

    nCalls = len(calls)
    print("Need to issue {} calls to Rucio".format(nCalls))
    callsDone = 0
    for call in calls:
        func = getattr(rucio, call[0])
        if len(call) > 1:
            signature = '%s:%s' % (call[0], sorted(viewitems(call[1])))
            result = func(**call[1])
        else:
            result = func()
            signature = call[0]
        callsDone += 1
        if callsDone % 10 == 0:
            print("Performed {} out of {} calls".format(callsDone, nCalls))
        finalResult.update({signature: result})

    print("Done!\n\nWriting out {} file".format(outFilename))
    with open(outFilename, 'w') as fileObj:
        json.dump(finalResult, fileObj, indent=1, separators=(',', ': '), sort_keys=True)


if __name__ == '__main__':
    sys.exit(main())



from builtins import str
import os


def ckey():
    "Return user CA key either from proxy or userkey.pem"
    pair = getKeyCertFromEnv()
    return pair[0]


def cert():
    "Return user CA cert either from proxy or usercert.pem"
    pair = getKeyCertFromEnv()
    return pair[1]


def getKeyCertFromEnv():
    """
    gets key and certificate from environment variables
    If no env variable is set return None, None for key, cert tuple

    First preference to HOST Certificate, This is how it set in Tier0

    """
    envPairs = [('X509_HOST_KEY', 'X509_HOST_CERT'),  # First preference to HOST Certificate,
                ('X509_USER_PROXY', 'X509_USER_PROXY'),  # Second preference to User Proxy, very common
                ('X509_USER_KEY', 'X509_USER_CERT')]  # Third preference to User Cert/Proxy combinition

    for keyEnv, certEnv in envPairs:
        localKey = os.environ.get(keyEnv)
        localCert = os.environ.get(certEnv)
        if localKey and localCert and os.path.exists(localKey) and os.path.exists(localCert):
            # if it is found in env return key, cert
            return localKey, localCert

    # TODO: only in linux, unix case, add other os case
    # look for proxy at default location /tmp/x509up_u$uid
    localKey = localCert = '/tmp/x509up_u' + str(os.getuid())
    if os.path.exists(localKey):
        return localKey, localCert

    # Finary look for globaus location
    if (os.environ.get('HOME') and
        os.path.exists(os.environ['HOME'] + '/.globus/usercert.pem')  and
        os.path.exists(os.environ['HOME'] + '/.globus/userkey.pem')):

        localKey = os.environ['HOME'] + '/.globus/userkey.pem'
        localCert = os.environ['HOME'] + '/.globus/usercert.pem'
        return localKey, localCert
    # couldn't find the key, cert files
    return None, None


def getCAPathFromEnv():
    """
    _getCAPathFromEnv_

    Return the path of the CA certificates. The check is loose in the pycurl_manager:
    is capath == None then the server identity is not verified. To enable this check
    you need to set either the X509_CERT_DIR variable or the cacert key of the request.
    """
    return os.environ.get("X509_CERT_DIR")



#!/usr/bin/env python
"""
_AddDupsToFileset_

MySQL implementation of Files.AddDupsToFileset
"""

import time

from WMCore.Database.DBFormatter import DBFormatter


class AddDupsToFileset(DBFormatter):
    existSQL = """SELECT lfn FROM wmbs_file_details wfd
                    INNER JOIN wmbs_fileset_files wff ON wff.fileid = wfd.id
                    WHERE wff.fileset = :fileset
                    AND wfd.lfn = :lfn"""

    sql = """INSERT IGNORE INTO wmbs_fileset_files (fileid, fileset, insert_time)
               SELECT wmbs_file_details.id, :fileset, :insert_time
               FROM wmbs_file_details
               WHERE wmbs_file_details.lfn = :lfn AND NOT EXISTS
                 (SELECT lfn FROM wmbs_file_details
                    INNER JOIN wmbs_fileset_files ON
                      wmbs_file_details.id = wmbs_fileset_files.fileid
                    INNER JOIN wmbs_subscription ON
                      wmbs_fileset_files.fileset = wmbs_subscription.fileset
                    INNER JOIN wmbs_workflow ON
                      wmbs_subscription.workflow = wmbs_workflow.id
                    WHERE wmbs_file_details.lfn = :lfn AND
                          wmbs_workflow.name = :workflow)"""

    sqlAvail = """INSERT IGNORE INTO wmbs_sub_files_available (subscription, fileid)
                    SELECT wmbs_subscription.id AS subscription,
                           wmbs_file_details.id AS fileid FROM wmbs_subscription
                      INNER JOIN wmbs_file_details ON
                        wmbs_file_details.lfn = :lfn
                    WHERE wmbs_subscription.fileset = :fileset AND NOT EXISTS
                 (SELECT lfn FROM wmbs_file_details
                    INNER JOIN wmbs_fileset_files ON
                      wmbs_file_details.id = wmbs_fileset_files.fileid
                    INNER JOIN wmbs_subscription ON
                      wmbs_fileset_files.fileset = wmbs_subscription.fileset
                    INNER JOIN wmbs_workflow ON
                      wmbs_subscription.workflow = wmbs_workflow.id
                    WHERE wmbs_file_details.lfn = :lfn AND
                          wmbs_workflow.name = :workflow AND
                          wmbs_fileset_files.fileset != :fileset)"""

    def execute(self, file, fileset, workflow, conn=None, transaction=False):
        binds = []
        availBinds = []
        existBinds = []
        timestamp = int(time.time())
        for fileLFN in file:
            existBinds.append({'lfn': fileLFN, 'fileset': fileset})
        existResult = self.dbi.processData(self.existSQL, existBinds,
                                           conn=conn,
                                           transaction=transaction)
        existsLFNs = self.formatDict(existResult)
        for result in existsLFNs:
            file.remove(result.get('lfn'))

        if len(file) < 1:
            # We removed all the files
            # There's nothing more for us to do.
            return

        for fileLFN in file:
            binds.append({"lfn": fileLFN, "fileset": fileset,
                          "insert_time": timestamp, "workflow": workflow})
            availBinds.append({"lfn": fileLFN, "fileset": fileset,
                               "workflow": workflow})

        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)
        self.dbi.processData(self.sqlAvail, availBinds, conn=conn,
                             transaction=transaction)
        return



#!/usr/bin/env python
"""
_CMSSWFetcher_

Fetch configfiles and PSet TWeaks for CMSSW Steps in a WMTask

"""

from future import standard_library
standard_library.install_aliases()

from WMCore.WMSpec.Steps.Fetchers.FetcherInterface import FetcherInterface
from WMCore.Cache.WMConfigCache import ConfigCache

import WMCore.WMSpec.WMStep as WMStep

class CMSSWFetcher(FetcherInterface):
    """
    _CMSSWFetcher_

    Pull configs from local config cache and add to sandbox.
    """
    def __call__(self, wmTask):
        """
        Trip through steps, find CMSSW steps, pull in config files,
        PSet Tweaks etc

        """
        for t in wmTask.steps().nodeIterator():
            t = WMStep.WMStepHelper(t)
            stepPath = "%s/%s" % (self.workingDirectory(), t.name())

            # the CMSSW has a special case with its ConfigCache argument
            if not t.stepType() in ("CMSSW"):
                continue
            if getattr(t.data.application.configuration, 'configCacheUrl', None) != None:
                # main config file
                fileTarget = "%s/%s" % (stepPath, t.data.application.command.configuration)
                #urllib.request.urlretrieve(
                #    t.data.application.configuration.retrieveConfigUrl,
                #    fileTarget)
                # PSet Tweak
                cacheUrl = t.data.application.configuration.configCacheUrl
                cacheDb  = t.data.application.configuration.cacheName
                configId = t.data.application.configuration.configId

                configCache = ConfigCache(cacheUrl, cacheDb)
                configCache.loadByID(configId)
                configCache.saveConfigToDisk(targetFile = fileTarget)



#!/usr/bin/env python
"""
_SubscriptionStatus_

Oracle implementation of Monitoring.SubscriptionStatus
"""




from WMCore.WMBS.MySQL.Monitoring.SubscriptionStatus import SubscriptionStatus \
 as SubscriptionStatusMySQL

class SubscriptionStatus(SubscriptionStatusMySQL):
    pass



#!/usr/bin/env python
"""
Unittests for IteratorTools functions
"""

from builtins import range
import itertools
import unittest

from Utils.IteratorTools import grouper, flattenList, makeListElementsUnique


class IteratorToolsTest(unittest.TestCase):
    """
    unittest for IteratorTools functions
    """

    def testGrouper(self):
        """
        Test the grouper function (returns chunk of an iterable)
        """

        listChunks = [i for i in grouper(list(range(0, 7)), 3)]  # Want list(range) for python 3
        iterChunks = [i for i in grouper(list(range(0, 7)), 3)]  # xrange becomes range in python 3

        for a, b in itertools.zip_longest(listChunks, iterChunks):
            self.assertEqual(a, b)

        self.assertEqual(listChunks[-1], [6])

    def testFlattenList(self):
        """
        Test the flattenList function (returns a flat list out
        of a list of lists)
        """
        doubleList = [list(range(1, 4)), list(range(10, 11)), list(range(15, 18))]
        flatList = flattenList(doubleList)
        self.assertEqual(len(flatList), 7)
        self.assertEqual(set(flatList), set([1, 2, 3, 10, 15, 16, 17]))

    def testNoDupListOfLists(self):
        """
        Test the `makeListElementsUnique` function (which returns a list with
        unique elements)
        """
        expRes = [[1], [2], [3], [4]]
        inputTest = [[1], [2], [3], [4]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        inputTest = [[1], [2], [3], [4], [1]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        inputTest = [[1], [2], [3], [4], [1], [1], [2], [3]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        # trying a different data type
        expRes = [[2, 20], [4, 40], [4, 41], [5, 40]]
        inputTest = [[2, 20], [4, 40], [4, 41], [5, 40], [4, 40], [4, 41]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        inputTest = [[5, 40], [2, 20], [4, 40], [4, 41], [5, 40], [4, 40], [4, 41]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        # and now with strings
        expRes = [["2", "20"], ["4", "40"], ["4", "41"], ["5", "40"]]
        inputTest = [["2", "20"], ["4", "40"], ["4", "41"], ["5", "40"], ["4", "40"], ["4", "41"]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        inputTest = [["5", "40"], ["2", "20"], ["4", "40"], ["4", "41"], ["5", "40"], ["4", "40"], ["4", "41"]]
        self.assertListEqual(expRes, makeListElementsUnique(inputTest))

        # now test with a list of tuples
        data = [(1, 2), (1, 3), (2, 4), (2, 5), (1, 3), (2, 5)]
        self.assertListEqual([(1, 2), (1, 3), (2, 4), (2, 5)], makeListElementsUnique(data))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
CouchService_t.py

Created by Dave Evans on 2010-10-04.
Copyright (c) 2010 Fermilab. All rights reserved.
"""
from builtins import range
import unittest
import random
import time

from WMCore.ACDC.CouchService import CouchService
from WMCore.ACDC.CouchFileset import CouchFileset
from WMCore.ACDC.CouchCollection import CouchCollection

from WMCore.Database.CMSCouch import CouchServer
from WMCore.DataStructs.File import File
from WMCore.Services.UUIDLib import makeUUID

from WMQuality.TestInitCouchApp import TestInitCouchApp


class CouchService_t(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Install the couch apps.
        """
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setupCouch("wmcore-acdc-couchservice", "GroupUser",
                                 "ACDC")
        return

    def tearDown(self):
        """
        _tearDown_

        Clean up couch.
        """
        self.testInit.tearDownCouch()
        return

    def populateCouchDB(self):
        """
        _populateCouchDB_

        Populate the ACDC records
        """
        testCollectionA = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="Thunderstruck")
        testCollectionB = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="Struckthunder")
        testCollectionC = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="Thunderstruck")
        testCollectionD = CouchCollection(database=self.testInit.couchDbName,
                                          url=self.testInit.couchUrl,
                                          name="Thunderstruck")

        testFilesetA = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetA")
        testCollectionA.addFileset(testFilesetA)
        testFilesetB = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetB")
        testCollectionB.addFileset(testFilesetB)
        testFilesetC = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetC")
        testCollectionC.addFileset(testFilesetC)
        testFilesetD = CouchFileset(database=self.testInit.couchDbName,
                                    url=self.testInit.couchUrl,
                                    name="TestFilesetD")
        testCollectionD.addFileset(testFilesetD)

        testFiles = []
        for i in range(5):
            testFile = File(lfn=makeUUID(), size=random.randint(1024, 4096),
                            events=random.randint(1024, 4096))
            testFiles.append(testFile)

        testFilesetA.add(testFiles)
        time.sleep(1)
        testFilesetB.add(testFiles)
        time.sleep(1)
        testFilesetC.add(testFiles)
        time.sleep(1)
        testFilesetD.add(testFiles)
        # Alan: unsure why to return this specific collection
        return testCollectionD

    def testListCollectionsFilesets(self):
        """
        _testListCollectionsFilesets_

        Verify that collections and filesets in ACDC can be listed.
        """
        testCollection = self.populateCouchDB()

        svc = CouchService(url=self.testInit.couchUrl,
                           database=self.testInit.couchDbName)

        goldenFilesetNames = ["TestFilesetA", "TestFilesetC", "TestFilesetD"]
        for fileset in svc.listFilesets(testCollection):
            self.assertTrue(fileset["name"] in goldenFilesetNames, "Error: Missing fileset.")
            goldenFilesetNames.remove(fileset["name"])
        self.assertEqual(len(goldenFilesetNames), 0, "Error: Missing filesets.")

        return

    def testTimestampAccounting(self):
        """
        _testTimestampAccounting_

        Check the correct functioning of the timestamp view in the ACDC
        couchapp and the function to remove old filesets.
        """
        self.populateCouchDB()
        svc = CouchService(url=self.testInit.couchUrl,
                           database=self.testInit.couchDbName)

        currentTime = time.time()
        database = CouchServer(self.testInit.couchUrl).connectDatabase(self.testInit.couchDbName)
        results = database.loadView("ACDC", "byTimestamp", {"endkey": currentTime})
        self.assertEqual(len(results["rows"]), 4)
        results = database.loadView("ACDC", "byTimestamp", {"endkey": currentTime - 2})
        self.assertEqual(len(results["rows"]), 2)
        results = database.loadView("ACDC", "byTimestamp", {"endkey": currentTime - 3})
        self.assertEqual(len(results["rows"]), 1)
        results = database.loadView("ACDC", "byTimestamp", {"endkey": currentTime - 5})
        self.assertEqual(len(results["rows"]), 0)
        svc.removeOldFilesets(0)
        results = database.loadView("ACDC", "byTimestamp", {"endkey": currentTime})
        self.assertEqual(len(results["rows"]), 0)
        return

    def testRemoveByCollectionName(self):
        """
        _testRemoveByCollectionName_

        Check the function to obliterate all the filesets of a collection
        """
        self.populateCouchDB()
        svc = CouchService(url=self.testInit.couchUrl,
                           database=self.testInit.couchDbName)
        database = CouchServer(self.testInit.couchUrl).connectDatabase(self.testInit.couchDbName)

        results = database.loadView("ACDC", "byCollectionName", keys=["Thunderstruck"],
                                    options={"reduce": False})
        self.assertTrue(len(results["rows"]) > 0)
        svc.removeFilesetsByCollectionName("Thunderstruck")
        results = database.loadView("ACDC", "byCollectionName", keys=["Thunderstruck"],
                                    options={"reduce": False})
        self.assertEqual(len(results["rows"]), 0)

        results = database.loadView("ACDC", "byCollectionName", keys=["Struckthunder"],
                                    options={"reduce": False})
        self.assertTrue(len(results["rows"]) > 0)
        svc.removeFilesetsByCollectionName("Struckthunder")
        results = database.loadView("ACDC", "byCollectionName", keys=["Struckthunder"],
                                    options={"reduce": False})
        self.assertEqual(len(results["rows"]), 0)
        return


if __name__ == '__main__':
    unittest.main()



"""
Unit tests for the WMCore/MicroService/DataStructs/NanoWorkflow module
"""
import unittest

from WMCore.MicroService.MSTransferor.RequestInfo import isNanoWorkflow, rsesIntersection


class RequestInfoTest(unittest.TestCase):
    """
    Test the very basic functionality of the RequestInfo module
    """

    def setUp(self):
        """
        Defined some basic data structs to use in the unit tests
        :return: None
        """
        pass

    def tearDown(self):
        pass

    def testIsNanoWorkflow(self):
        """
        Test object instance type
        """
        testDict = {'RequestType': 'ReReco'}
        self.assertFalse(isNanoWorkflow(testDict))
        testDict = {'RequestType': 'ReReco', 'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAODSIM'}
        self.assertFalse(isNanoWorkflow(testDict))

        testDict = {'RequestType': 'TaskChain', 'Task1': {}, 'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAODSIM'}
        self.assertFalse(isNanoWorkflow(testDict))
        testDict = {'RequestType': 'TaskChain', 'Task1': {'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAOD'}}
        self.assertFalse(isNanoWorkflow(testDict))
        testDict = {'RequestType': 'TaskChain', 'Task1': {'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAODSIM'}}
        self.assertTrue(isNanoWorkflow(testDict))

        testDict = {'RequestType': 'StepChain', 'Step1': {}, 'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAODSIM'}
        self.assertFalse(isNanoWorkflow(testDict))
        testDict = {'RequestType': 'StepChain', 'Step1': {'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAOD'}}
        self.assertFalse(isNanoWorkflow(testDict))
        testDict = {'RequestType': 'StepChain', 'Step1': {'InputDataset': '/PrimDset/AcqEra-ProcStr/MINIAODSIM'}}
        self.assertTrue(isNanoWorkflow(testDict))

    def testrsesIntersection(self):
        """
        Test the rsesIntersection function
        """
        rses = []
        self.assertEqual(rses, rsesIntersection(rses))

        rses = [['aaa', 'bbb', 'ccc']]
        self.assertEqual(rses[0], rsesIntersection(rses))

        rses = [['aaa', 'bbb', 'ccc'], ['bbb', 'ccc'], ['ccc']]
        self.assertEqual(rses[2], rsesIntersection(rses))

        rses = [['aaa', 'bbb', 'ccc'], ['ddd']]
        self.assertEqual([], rsesIntersection(rses))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
Oracle implementation of File.SetParentageByMergeJob

Make the parentage link between a file and all the inputs of a given job
"""




from WMCore.WMBS.MySQL.Files.SetParentageByMergeJob import SetParentageByMergeJob as MySQLSetParentageByMergeJob

class SetParentageByMergeJob(MySQLSetParentageByMergeJob):
    pass



"""
_ListSubscriptions_

MySQL implementation of DBS3Buffer.ListSubscriptions

Created on May 2, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListSubscriptions(DBFormatter):
    """
    _ListSubscriptions_

    List the associated subscriptions to a dataset given
    its dataset id
    """

    sql = """SELECT *
             FROM dbsbuffer_dataset_subscription
             WHERE dataset_id = :id
          """


    sqlWithSub = """SELECT *
                    FROM dbsbuffer_dataset_subscription
                    WHERE dataset_id = :id AND subscribed = :subscribed
                 """

    def execute(self, datasetID, subscribed = None,
                conn = None,
                transaction = False):
        """
        _execute_

        Execute the query and retrieve the subscriptions
        """
        if subscribed is not None:
            result = self.dbi.processData(self.sql,
                                          {'id' : datasetID,
                                           'subscribed' : subscribed}, conn = conn,
                                          transaction = transaction)
        else:
            result = self.dbi.processData(self.sql,
                                          {'id' : datasetID}, conn = conn,
                                          transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_DBSBufferUtil_t_

Unit tests for DBSBufferUtil class
"""

from builtins import range
import unittest
import threading

from WMCore.DAOFactory import DAOFactory
from WMQuality.TestInit import TestInit
from WMCore.WMBS.Workflow import Workflow
from WMComponent.DBS3Buffer.DBSBufferUtil import DBSBufferUtil
from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile

class DBSBufferUtilTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Create some DBSBuffer tables and fake data for testing
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMComponent.DBS3Buffer", "WMCore.WMBS"],
                                useDefault=False)

        myThread = threading.currentThread()
        self.dbsbufferFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        self.wmbsFactory = DAOFactory(package="WMCore.WMBS",
                                           logger=myThread.logger,
                                           dbinterface=myThread.dbi)
        self.dbsUtil = DBSBufferUtil()

        # Create two test dbsbuffer workflows
        insertWorkflow = self.dbsbufferFactory(classname="InsertWorkflow")
        insertWorkflow.execute("Test1", "Task1", 0, 0, 0, 0)
        insertWorkflow.execute("Test2", "Task2", 0, 0, 0, 0)

        # Update one workflow to "completed" state
        updateWorkflow = self.dbsbufferFactory(classname="UpdateWorkflowsToCompleted")
        updateWorkflow.execute(["Test1"])

        # Create a test wmbs workflow
        testWorkflow = Workflow(spec="somespec.xml", owner="Erik", name="Test1", task="Task1")
        testWorkflow.create()

        # Create a test dbsbuffer file
        self.createTestFiles()

    def createTestFiles(self):
        """
        _createTestFiles_

        Create some dbsbuffer test files with different statuses
        :return:
        """
        phedexStatus = self.dbsbufferFactory(classname="DBSBufferFiles.SetPhEDExStatus")

        for i in range(0, 4):

            lfn = "/path/to/some/lfn" + str(i)

            # Two files should be InDBS, two files should be NOTUPLOADED
            if i in [0,2]:
                status = 'InDBS'
            else:
                status = 'NOTUPLOADED'

            testDBSFile = DBSBufferFile(lfn=lfn, size=600000, events=60000, status=status, workflowId=1)

            testDBSFile.setAlgorithm(appName="cmsRun", appVer="UNKNOWN",
                                     appFam="RECO", psetHash="SOMEHASH" + str(i),
                                     configContent="SOMECONTENT")
            testDBSFile.setDatasetPath("/path/to/some/dataset")
            testDBSFile.create()

            # Create all four combinations of status(InDBS,NOTUPLOADED) and in_phedex(0,1)
            if i in [0,1]:
                phedexStatus.execute(lfn, 1)

    def tearDown(self):
        """
        _tearDown_

        Drop all the DBSBuffer tables.
        """
        self.testInit.clearDatabase()

    # List of methods to potentially test
    # def loadDBSBufferFilesBulk(self, fileObjs):
    # def findUploadableDAS(self):
    # def testFindOpenBlocks(self):
    # def loadBlocksByDAS(self, das):
    #
    # def loadBlocks(self, blocknames):
    #
    # def findUploadableFilesByDAS(self, datasetpath):
    #
    # def loadFilesByBlock(self, blockname):

    def testGetPhEDExDBSStatusForCompletedWorkflows(self):
        """
        _testGetPhEDExDBSStatusForCompletedWorkflows_

        :return:
        """

        results = self.dbsUtil.getPhEDExDBSStatusForCompletedWorkflows()
        self.assertEqual(results["Test1"]["InDBS"], 2, "ERROR: Files with InDBS status is incorrect.")
        self.assertEqual(results["Test1"]["InPhEDEx"], 2, "ERROR: Files with InPhEDEx status is incorrect.")
        self.assertEqual(results["Test1"]["NotInDBS"], 2, "ERROR: Files with NotInDBS status is incorrect.")
        self.assertEqual(results["Test1"]["NotInPhEDEx"], 2, "ERROR: Files with NotInPhEDEx status is incorrect.")

        return

    def testGetCompletedWorkflows(self):
        """
        _testGetCompletedWorkflows_

        :return:
        """

        results = self.dbsUtil.getCompletedWorkflows()
        self.assertEqual(len(results), 1, "ERROR: GetCompletedWorkflows returned incorrect number of completed workflows.")
        self.assertIn("Test1", results, "ERROR: GetCompletedWorkflows returned incorrect workflow.")

        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_AddValidation_

Oracle implementation of Subscription.AddValidation
"""

from WMCore.WMBS.MySQL.Subscriptions.AddValidation import AddValidation as AddValidationMySQL

class AddValidation(AddValidationMySQL):
    sql = """INSERT INTO wmbs_subscription_validation
                               (subscription_id, location_id, valid)
               SELECT :sub, id, :valid FROM wmbs_location
               WHERE site_name = :site_name AND NOT EXISTS
               (SELECT * FROM wmbs_subscription_validation
                WHERE subscription_id = :sub AND location_id = wmbs_location.id
                      AND valid = :valid)"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
_DBFormatterTest_

Unit tests for the DBFormatter class

"""
from __future__ import print_function

import threading
import unittest

from builtins import str

from WMCore.Database.DBFormatter import DBFormatter
from WMQuality.TestInit import TestInit


class DBFormatterTest(unittest.TestCase):
    """
    _DBFormatterTest_

    Unit tests for the DBFormatter class

    """

    def setUp(self):
        "make a logger instance and create tables"

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase=True)
        self.testInit.setSchema(customModules=["WMQuality.TestDB"],
                                useDefault=False)
        self.selectSQL = "SELECT * FROM test_tableb"

    def tearDown(self):
        """
        Delete the databases
        """
        self.testInit.clearDatabase()

    def stuffDB(self):
        """Populate one of the test tables"""
        insertSQL = "INSERT INTO test_tableb (column1, column2, column3) values (:bind1, :bind2, :bind3)"
        insertBinds = [{'bind1': u'value1a', 'bind2': 1, 'bind3': u'value2a'},
                       {'bind1': 'value1b', 'bind2': 2, 'bind3': 'value2b'},
                       {'bind1': b'value1c', 'bind2': 3, 'bind3': b'value2d'}]

        myThread = threading.currentThread()
        myThread.dbi.processData(insertSQL, insertBinds)

    def testBFormatting(self):
        """
        Test various formats
        """
        # fill the database with some initial data
        self.stuffDB()

        myThread = threading.currentThread()
        dbformatter = DBFormatter(myThread.logger, myThread.dbi)

        result = myThread.dbi.processData(self.selectSQL)
        output = dbformatter.format(result)
        self.assertEqual(output, [['value1a', 1, 'value2a'],
                                  ['value1b', 2, 'value2b'],
                                  ['value1c', 3, 'value2d']])

        result = myThread.dbi.processData(self.selectSQL)
        output = dbformatter.formatOne(result)
        print('test1 ' + str(output))
        self.assertEqual(output, ['value1a', 1, 'value2a'])

        result = myThread.dbi.processData(self.selectSQL)
        output = dbformatter.formatDict(result)
        self.assertEqual(output, [{'column3': 'value2a', 'column2': 1, 'column1': 'value1a'},
                                  {'column3': 'value2b', 'column2': 2, 'column1': 'value1b'},
                                  {'column3': 'value2d', 'column2': 3, 'column1': 'value1c'}])

        result = myThread.dbi.processData(self.selectSQL)
        output = dbformatter.formatOneDict(result)
        self.assertEqual(output, {'column3': 'value2a', 'column2': 1, 'column1': 'value1a'})


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from future.utils import viewitems

import hashlib
import hmac

import cherrypy

from Utils.Utilities import lowerCmsHeaders


def get_user_info():
    """
    Helper function to return user based information of the request
    """
    return cherrypy.request.user


class FrontEndAuth(cherrypy.Tool):
    """
    Transparently allows a back-end cmsweb app to do
    authn/z based on the headers sent by the front-end.
    """

    def __init__(self, config):
        """Read hmac secret  and define cherrypy hook point."""
        # reads the bin key used to verify the hmac
        with open(config.key_file, "rb") as f:
            self.key = f.read()

        # Defines the hook point for cherrypy
        self._name = None
        self._point = 'before_request_body'
        self._priority = 60  # Just after the sessions being enabled

    def callable(self, role=None, group=None, site=None, authzfunc=None):
        """
        This is the method that is called in the cherrypy hook point (whenever
        the user-agent requests a page of the back-end app.
        """
        role = role or []
        group = group or []
        site = site or []
        # Sets initial user information for this request
        assert getattr(cherrypy.request, "user", None) is None
        cherrypy.request.user = {'dn': None,
                                 'method': None,
                                 'login': None,
                                 'name': None,
                                 'roles': {}}

        # Checks authn by reading front-end headers
        self.check_authentication()

        # Now checks authz
        self.check_authorization(role, group, site, authzfunc)

        # The user is authenticated and authorized. Then cherrypy will
        # call the proper app handler to show the requested page.
        # User info still available under cherrypy.thread_data.user.

    def check_authentication(self):
        """Read and verify the front-end headers, update the user
        dict with information about the authorized user."""
        headers = lowerCmsHeaders(cherrypy.request.headers)
        user = get_user_info()

        if 'cms-auth-status' not in headers:
            # Non SSL request
            raise cherrypy.HTTPError(403, "You are not allowed to access this resource.")

        if headers['cms-auth-status'] == 'NONE':
            # User authentication is optional
            return  # authn accepted

        # User information is available on headers
        prefix = suffix = ""
        hkeys = sorted(headers.keys())
        for hk in hkeys:
            hk = hk.lower()
            if hk[0:9] in ["cms-authn", "cms-authz"] and hk != "cms-authn-hmac":
                prefix += "h%xv%x" % (len(hk), len(headers[hk]))
                suffix += "%s%s" % (hk, headers[hk])
                hkname = hk.split('-', 2)[-1]
                if hk.startswith("cms-authn"):
                    user[hkname] = headers[hk]
                if hk.startswith("cms-authz"):
                    user['roles'][hkname] = {'site': set(), 'group': set()}
                    for r in headers[hk].split():
                        ste_or_grp, name = r.split(':')
                        user['roles'][hkname][ste_or_grp].add(name)

        vfy = hmac.new(self.key, prefix + "#" + suffix, hashlib.sha1).hexdigest()
        if vfy != headers["cms-authn-hmac"]:
            # HMAC does not match
            raise cherrypy.HTTPError(403, "You are not allowed to access this resource, hmac mismatch")

            # User authn accepted

    def check_authorization(self, role, group, site, authzfunc):
        """Format the authorization rules into lists and verify if the given
        user is allowed to access."""
        if authzfunc is None:
            authzfunc = self.defaultAuth

        # TOFIX: put role, group and site into canonical form

        # Turns arguments into lists
        if role and isinstance(role, str):
            role = [role]
        if group and isinstance(group, str):
            group = [group]
        if site and isinstance(site, str):
            site = [site]

        # Finally checks if the user is allowed
        if not authzfunc(get_user_info(), role, group, site):
            # Authorization denied
            raise cherrypy.HTTPError(403, "You are not allowed to access this resource, authz denied")

    def defaultAuth(self, user, role, group, site):
        """ Return True for authorized user, False otherwise.

        An user is authorized if he has any of the asked roles in the
        given sites or groups. When no roles is specified, belonging
        to any of the asked sites or groups is enough.
        """
        if not (role or group or site):
            return True

        for k, v in viewitems(user['roles']):
            if (not role) or (k in role):
                if not (group or site):
                    return True
                if set(group) & v['group']:
                    return True
                if set(site) & v['site']:
                    return True
        return False


class NullAuth(cherrypy.Tool):
    def __init__(self, config):
        # Defines the hook point for cherrypy
        self._name = None
        self._point = 'before_request_body'
        self._priority = 60  # Just after the sessions being enabled
        if cherrypy.server.environment == 'production':
            cherrypy.log.access_log.critical('You MUST NOT use the NullAuth in a production environment')
            raise cherrypy.CherryPyException('Invalid server authentication')
        else:
            cherrypy.log.access_log.warning("You are using the NullAuth, I hope you know what you're doing")

    def callable(self, role=None, group=None, site=None, authzfunc=None):
        role = role or []
        group = group or []
        site = site or []
        cherrypy.log.access_log.warning('NullAuth called for:')
        cherrypy.log.access_log.warning('\trole(s): %s \n\tgroup(s): %s \n\tsite(s): %s', role, group, site)

        if authzfunc:
            cherrypy.log.access_log.warning('\tusing authorisation function %s', authzfunc.__name__)
        cherrypy.request.user = {'dn': 'None',
                                 'method': 'Null Auth - totally insecure!',
                                 'login': 'fbloggs',
                                 'name': 'Fred Bloggs',
                                 'roles': {}}



#!/usr/bin/env python
"""
_GetJobGroups_

MySQL implementation of Subscriptions.GetJobGroups
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetJobGroups(DBFormatter):
    sql = """SELECT DISTINCT wmbs_jobgroup.id FROM wmbs_jobgroup
               INNER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
              WHERE wmbs_jobgroup.subscription = :subscription AND
                    wmbs_job_state.name = 'new'"""

    def format(self, results):
        """
        _format_

        Format the results into a single list of job group IDs.
        """
        results = DBFormatter.format(self, results)

        jobGroupIDs = []
        for result in results:
            for row in result:
                jobGroupIDs.append(int(row))

        return jobGroupIDs

    def execute(self, subscription = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {"subscription": subscription},
                                      conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_TestInit__

A set of initialization steps used in many tests.
Test can call the methods from this class to
initialize their default environment so to
minimize code duplication.

This class is not a test but an auxilary class and
is based on the WMCore.WMInit class.

"""
from __future__ import print_function

from builtins import object

import logging
import os
import shutil
import tempfile
import threading
import traceback

from WMCore.Agent.Configuration import Configuration
from WMCore.Agent.Configuration import loadConfigurationFile
from WMCore.WMException import WMException

hasDatabase = True
try:
    from WMCore.WMInit import WMInit
except ImportError as ex:
    print(str(ex))
    print("NOTE: TestInit is being loaded without database support")
    hasDatabase = False

# Sorry for the global, but I think this should go here
trashDatabases = False  # delete databases after every test?

class TestInitException(WMException):
    """
    _TestInitException_

    You should see this only if something is wrong setting up the database connection
    Still, just in case...
    """

def deleteDatabaseAfterEveryTest(areYouSerious):
    """
    this method handles whether or not TestInit will be vicious to databases
    """
    # python is idiotic for its scoping system
    global trashDatabases
    if areYouSerious == "I'm Serious":
        print("We are going to trash databases after every test")
        trashDatabases = True
    else:
        #"I'm glad you weren't serious"
        print("We are not going to trash databases after every test")
        trashDatabases = False


class TestInit(object):
    """
    A set of initialization steps used in many tests.
    Test can call the methods from this class to
    initialize their default environment so to
    minimize code duplication.
    """

    def __init__(self, testClassName = "Unknown Class"):
        self.testClassName = testClassName
        self.testDir = None
        self.currModules = []
        global hasDatabase
        self.hasDatabase = hasDatabase
        if self.hasDatabase:
            self.init = WMInit()
        self.deleteTmp = True

    def __del__(self):
        if self.deleteTmp:
            self.delWorkDir()
        self.attemptToCloseDBConnections()


    def delWorkDir(self):
        if self.testDir != None:
            try:
                shutil.rmtree( self.testDir )
            except:
                # meh, if it fails, I guess something weird happened
                pass

    def setLogging(self, logLevel = logging.INFO):
        """
        Sets logging parameters
        """
        # remove old logging instances.
        return
        logger1 = logging.getLogger()
        logger2 = logging.getLogger(self.testClassName)
        for logger in [logger1, logger2]:
            for handler in logger.handlers:
                handler.close()
                logger.removeHandler(handler)

        self.init.setLogging(self.testClassName, self.testClassName,
                             logExists = False, logLevel = logLevel)

    def generateWorkDir(self, config = None, deleteOnDestruction = True, setTmpDir = False):
        self.testDir = tempfile.mkdtemp()
        if config:
            config.section_("General")
            config.General.workDir = self.testDir
        os.environ['TESTDIR'] = self.testDir
        if os.getenv('WMCORE_KEEP_DIRECTORIES', False):
            deleteOnDestruction = True
            logging.info("Generated testDir - %s" % self.testDir)
        if setTmpDir:
            os.environ['TMPDIR'] = self.testDir

        self.deleteTmp = deleteOnDestruction
        return self.testDir

    def getBackendFromDbURL(self, dburl):
        dialectPart = dburl.split(":")[0]
        if dialectPart == 'mysql':
            return 'MySQL'
        elif dialectPart == 'oracle':
            return 'Oracle'
        elif dialectPart == 'http':
            return 'CouchDB'
        else:
            raise RuntimeError("Unrecognized dialect %s" % dialectPart)

    def setDatabaseConnection(self, connectUrl=None, socket=None, destroyAllDatabase = False):
        """
        Set up the database connection by retrieving the environment
        parameters.

        The destroyAllDatabase option is for testing ONLY.  Never flip that switch
        on in any other instance where you don't know what you're doing.
        """
        if not self.hasDatabase:
            return
        config = self.getConfiguration(connectUrl=connectUrl, socket=socket)
        self.coreConfig = config
        self.init.setDatabaseConnection(config.CoreDatabase.connectUrl,
                                        config.CoreDatabase.dialect,
                                        config.CoreDatabase.socket)

        if trashDatabases or destroyAllDatabase:
            self.clearDatabase()

        # Have to check whether or not database is empty
        # If the database is not empty when we go to set the schema, abort!
        try:
            result = self.init.checkDatabaseContents()
        except Exception as e:
            logging.debug("Error checking DB contents, assume DB does not exist")
            logging.debug(str(e))
            return
        if len(result) > 0:
            msg = "Database not empty, cannot set schema !\n"
            msg += str(result)
            logging.error(msg)
            raise TestInitException(msg)

        return

    def setSchema(self, customModules = [], useDefault = True, params = None):
        """
        Creates the schema in the database for the default
        tables/services: trigger, message service, threadpool.

        Developers can add their own modules to it using the array
        customModules which should follow the proper naming convention.

        if useDefault is set to False, it will not instantiate the
        schemas in the defaultModules array.
        """
        if not self.hasDatabase:
            return
        defaultModules = ["WMCore.WMBS"]
        if not useDefault:
            defaultModules = []

        # filter for unique modules
        modules = []
        for module in (defaultModules + customModules):
            if module not in modules:
                modules.append(module)

        try:
            self.init.setSchema(modules, params = params)
        except Exception as ex:
            print(traceback.format_exc())
            raise ex

        # store the list of modules we've added to the DB
        # again filter for unique modules
        modules = []
        for module in (defaultModules + customModules + self.currModules):
            if module not in modules:
                modules.append(module)
        self.currModules = modules

        return

    def getDBInterface(self):
        "shouldbe called after connection is made"
        if not self.hasDatabase:
            return
        myThread = threading.currentThread()

        return myThread.dbi

    def getConfiguration(self, configurationFile = None, connectUrl = None, socket=None):
        """
        Loads (if available) your configuration file and augments
        it with the standard settings used in multiple tests.
        """
        if configurationFile != None:
            config = loadConfigurationFile(configurationFile)
        else:
            config = Configuration()

        # some general settings that would come from the general default
        # config file
        config.Agent.contact = "fvlingen@caltech.edu"
        config.Agent.teamName = "Lakers"
        config.Agent.agentName = "Lebron James"
        config.Agent.hostName = "testhost.laker.world"

        config.section_("General")
        # If you need a testDir, call testInit.generateWorkDir
        # config.General.workDir = os.getenv("TESTDIR")
        config.General.ReqMgr2ServiceURL = "http://localhost/reqmgr2"

        config.section_("CoreDatabase")
        if connectUrl:
            config.CoreDatabase.connectUrl = connectUrl
            config.CoreDatabase.dialect = self.getBackendFromDbURL(connectUrl)
            config.CoreDatabase.socket = socket or os.getenv("DBSOCK")
        else:
            if os.getenv('DATABASE') == None:
                raise RuntimeError("You must set the DATABASE environment variable to run tests")
            config.CoreDatabase.connectUrl = os.getenv("DATABASE")
            config.CoreDatabase.dialect = self.getBackendFromDbURL(os.getenv("DATABASE"))
            config.CoreDatabase.socket = os.getenv("DBSOCK")
            if os.getenv("DBHOST"):
                print("****WARNING: the DBHOST environment variable will be deprecated soon***")
                print("****WARNING: UPDATE YOUR ENVIRONMENT OR TESTS WILL FAIL****")
            # after this you can augment it with whatever you need.

        couchurl = os.getenv("COUCHURL")
        config.section_("ACDC")
        config.ACDC.couchurl = couchurl
        config.ACDC.database = "wmagent_acdc_t"

        config.component_("JobStateMachine")
        config.JobStateMachine.couchurl = couchurl
        config.JobStateMachine.couchDBName = "wmagent_job_test"
        config.JobStateMachine.jobSummaryDBName = "job_summary"
        config.JobStateMachine.summaryStatsDBName = "stat_summary_test"

        config.component_("JobAccountant")
        config.JobAccountant.pollInterval = 60
        config.JobAccountant.componentDir = os.getcwd()
        config.JobAccountant.logLevel = 'SQLDEBUG'

        config.component_("TaskArchiver")
        config.TaskArchiver.localWMStatsURL = "%s/%s" % (config.JobStateMachine.couchurl, config.JobStateMachine.jobSummaryDBName)

        return config

    def clearDatabase(self, modules = []):
        """
        Database deletion. Global, ignore modules.
        """
        if not self.hasDatabase:
            return

        self.init.clearDatabase()

        return

    def attemptToCloseDBConnections(self):
        return
        myThread = threading.currentThread()
        print("Closing DB")

        try:
            if not myThread.transaction \
                 and not myThread.transaction.conn \
                 and not myThread.transaction.conn.closed:

                myThread.transaction.conn.close()
                myThread.transaction.conn = None
                print("Connection Closed")
        except Exception as e:
            print("tried to close DBI but failed: %s" % e)

        try:
            if hasattr(myThread, "dbFactory"):
                del myThread.dbFactory
                print("dbFactory removed")
        except Exception as e:
            print("tried to delete factory but failed %s" % e)



#!/usr/bin/env python
"""
ErrorHandler test TestErrorHandler module and the harness
"""
from __future__ import print_function

from builtins import range
import cProfile
import os
import os.path
import pstats
import threading
import time
import unittest

from Utils.PythonVersion import PY3

from WMCore_t.WMSpec_t.TestSpec import createTestWorkload
from nose.plugins.attrib import attr

import WMCore.WMBase
from WMComponent.ErrorHandler.ErrorHandlerPoller import ErrorHandlerPoller
from WMCore.ACDC.DataCollectionService import DataCollectionService
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.TestInitCouchApp import TestInitCouchApp


class ErrorHandlerTest(EmulatedUnitTestCase):
    """
    TestCase for TestErrorHandler module
    """

    def setUp(self):
        """
        setup for test.
        """
        super(ErrorHandlerTest, self).setUp()
        myThread = threading.currentThread()

        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)
        self.testInit.setupCouch("errorhandler_t", "GroupUser", "ACDC")
        self.testInit.setupCouch("errorhandler_t_jd/jobs", "JobDump")
        self.testInit.setupCouch("errorhandler_t_jd/fwjrs", "FWJRDump")

        self.daofactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.getJobs = self.daofactory(classname="Jobs.GetAllJobs")
        self.setJobTime = self.daofactory(classname="Jobs.SetStateTime")
        locationAction = self.daofactory(classname="Locations.New")
        locationAction.execute(siteName="malpaquet", pnn="T2_CH_CERN")
        self.testDir = self.testInit.generateWorkDir()
        self.configFile = EmulatorSetup.setupWMAgentConfig()
        self.nJobs = 10

        self.dataCS = DataCollectionService(url=self.testInit.couchUrl,
                                            database="errorhandler_t")

        if PY3:
            self.assertItemsEqual = self.assertCountEqual

        return

    def tearDown(self):
        """
        Database deletion
        """
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        EmulatorSetup.deleteConfig(self.configFile)
        return

    def getConfig(self):
        """
        _getConfig_

        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        # First the general stuff
        config.section_("General")
        config.General.workDir = os.getenv("TESTDIR", self.testDir)
        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        config.component_("ErrorHandler")
        # The log level of the component.
        config.ErrorHandler.logLevel = 'INFO'
        # The namespace of the component
        config.ErrorHandler.namespace = 'WMComponent.ErrorHandler.ErrorHandler'
        # maximum number of threads we want to deal
        # config.ErrorHandler.maxThreads = 30
        # with messages per pool.
        config.ErrorHandler.maxProcessSize = 30
        config.ErrorHandler.readFWJR = True
        # maximum number of retries we want for job
        config.ErrorHandler.maxRetries = 5
        # The poll interval at which to look for failed jobs
        config.ErrorHandler.pollInterval = 60
        # this will be overwritten in some unittests
        # JobStateMachine
        config.component_('JobStateMachine')
        config.JobStateMachine.couchurl = os.getenv('COUCHURL', None)
        config.JobStateMachine.couchDBName = "errorhandler_t_jd"

        config.section_('ACDC')
        config.ACDC.couchurl = self.testInit.couchUrl
        config.ACDC.database = "errorhandler_t"

        return config

    def createWorkload(self, workloadName='Test'):
        """
        Creates a test workload for us to run on, hold the basic necessities.
        """

        workload = createTestWorkload(workloadName)

        # Add RequestManager stuff
        workload.data.request.section_('schema')
        workload.data.request.schema.Requestor = 'nobody'
        workload.data.request.schema.Group = 'testers'

        taskMaker = TaskMaker(workload, os.path.join(self.testDir, 'workloadTest'))
        taskMaker.skipSubscription = True
        taskMaker.processWorkload()

        return workload

    def createTestJobGroup(self, nJobs=10, retry_count=1,
                           workloadPath='test', fwjrPath=None,
                           workloadName=makeUUID(),
                           fileModifier=''):
        """
        Creates a group of several jobs
        """

        myThread = threading.currentThread()
        myThread.transaction.begin()
        testWorkflow = Workflow(spec=workloadPath, owner="cmsdataops", group="cmsdataops",
                                name=workloadName, task="/TestWorkload/ReReco")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFile0 = File(lfn="/this/is/a/parent%s" % fileModifier, size=1024, events=10)
        testFile0.addRun(Run(10, *[12312]))
        testFile0.setLocation('T2_CH_CERN')

        testFileA = File(lfn="/this/is/a/lfnA%s" % fileModifier, size=1024, events=10,
                         first_event=88, merged=False)
        testFileA.addRun(Run(10, *[12312, 12313]))
        testFileA.setLocation('T2_CH_CERN')

        testFileB = File(lfn="/this/is/a/lfnB%s" % fileModifier, size=1024, events=10,
                         first_event=88, merged=False)
        testFileB.addRun(Run(10, *[12314, 12315, 12316]))
        testFileB.setLocation('T2_CH_CERN')

        testFile0.create()
        testFileA.create()
        testFileB.create()

        testFileA.addParent(lfn="/this/is/a/parent%s" % fileModifier)
        testFileB.addParent(lfn="/this/is/a/parent%s" % fileModifier)

        for i in range(0, nJobs):
            testJob = Job(name=makeUUID())
            testJob['retry_count'] = retry_count
            testJob['retry_max'] = 10
            testJob['mask'].addRunAndLumis(run=10, lumis=[12312])
            testJob['mask'].addRunAndLumis(run=10, lumis=[12314, 12316])
            testJob['cache_dir'] = os.path.join(self.testDir, testJob['name'])
            testJob['fwjr_path'] = fwjrPath
            os.mkdir(testJob['cache_dir'])
            testJobGroup.add(testJob)
            testJob.create(group=testJobGroup)
            testJob.addFile(testFileA)
            testJob.addFile(testFileB)
            testJob.save()

        testJobGroup.commit()

        testSubscription.acquireFiles(files=[testFileA, testFileB])
        testSubscription.save()
        myThread.transaction.commit()

        return testJobGroup

    def testA_Create(self):
        """
        WMComponent_t.ErrorHandler_t.ErrorHandler_t:testCreate()

        Mimics creation of component and test jobs failed in create stage.
        """
        njobs = 4
        workloadName = 'TestWorkload'

        self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', workloadName,
                                    'WMSandbox', 'WMWorkload.pkl')

        #        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs,
        testJobGroup = self.createTestJobGroup(nJobs=njobs,
                                               workloadPath=workloadPath,
                                               workloadName=workloadName)
        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'createfailed', 'created')

        idList = self.getJobs.execute(state='CreateFailed')
        self.assertEqual(len(idList), njobs)

        testErrorHandler = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler.reqAuxDB = None
        testErrorHandler.setup(None)
        testErrorHandler.algorithm(None)

        idList = self.getJobs.execute(state='CreateFailed')
        self.assertEqual(len(idList), 0)

        # These should go directly to exhausted
        idList = self.getJobs.execute(state='Exhausted')
        self.assertEqual(len(idList), njobs)

        # Check that it showed up in ACDC
        collection = self.dataCS.getDataCollection(workloadName)

        # Now look at what's inside
        self.assertTrue(len(collection['filesets']) > 0)
        for fileset in collection["filesets"]:
            counter = 0
            for f in fileset.listFiles():
                counter += 1
                self.assertTrue(f['lfn'] in ["/this/is/a/lfnA", "/this/is/a/lfnB"])
                self.assertEqual(f['events'], 10)
                self.assertEqual(f['size'], 1024)
                self.assertEqual(f['parents'], ['/this/is/a/parent'])
                self.assertTrue(f['runs'][0]['run_number'] == 10)
                if f['lfn'] == "/this/is/a/lfnA":
                    self.assertItemsEqual(f['runs'][0]['lumis'], [12312])
                elif f['lfn'] == "/this/is/a/lfnB":
                    self.assertItemsEqual(f['runs'][0]['lumis'], [12314, 12315, 12316])
                else:
                    self.assertFail("File name is not known: %s" % f['lfn'])
                self.assertEqual(f['merged'], 0)
                self.assertEqual(f['first_event'], 88)
            self.assertEqual(counter, njobs * 2)  # each job has 2 files (thus 4 times duplicate)
        return

    def testB_Submit(self):
        """
        WMComponent_t.ErrorHandler_t.ErrorHandler_t:testSubmit()

        Mimics creation of component and test jobs failed in submit stage.
        """
        workloadName = 'TestWorkload'

        self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', workloadName,
                                    'WMSandbox', 'WMWorkload.pkl')

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs,
                                               workloadPath=workloadPath)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'submitfailed', 'created')

        idList = self.getJobs.execute(state='SubmitFailed')
        self.assertEqual(len(idList), self.nJobs)

        testErrorHandler = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler.reqAuxDB = None
        testErrorHandler.setup(None)
        testErrorHandler.algorithm(None)

        idList = self.getJobs.execute(state='SubmitFailed')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='SubmitCooloff')
        self.assertEqual(len(idList), self.nJobs)
        return

    def testC_Jobs(self):
        """
        WMComponent_t.ErrorHandler_t.ErrorHandler_t.testJobs()

        Mimics creation of component and test jobs failed in execute stage.
        """
        workloadName = 'TestWorkload'

        self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', workloadName,
                                    'WMSandbox', 'WMWorkload.pkl')

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs,
                                               workloadPath=workloadPath)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')

        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), self.nJobs)

        testErrorHandler = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler.reqAuxDB = None
        testErrorHandler.setup(None)
        testErrorHandler.algorithm(None)

        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), self.nJobs)
        return

    def testD_Exhausted(self):
        """
        _testExhausted_

        Test that the system can exhaust jobs correctly
        """
        workloadName = 'TestWorkload'

        self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', workloadName,
                                    'WMSandbox', 'WMWorkload.pkl')

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs, retry_count=5,
                                               workloadPath=workloadPath)

        config = self.getConfig()
        config.ErrorHandler.maxRetries = 1
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')

        testSubscription = Subscription(id=1)  # You should only have one
        testSubscription.load()
        testSubscription.loadData()

        # Do we have files to start with?
        self.assertEqual(len(testSubscription.filesOfStatus("Acquired")), 2)

        testErrorHandler = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler.reqAuxDB = None
        testErrorHandler.setup(None)
        testErrorHandler.algorithm(None)

        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='Exhausted')
        self.assertEqual(len(idList), self.nJobs)

        # Did we fail the files?
        self.assertEqual(len(testSubscription.filesOfStatus("Acquired")), 0)
        self.assertEqual(len(testSubscription.filesOfStatus("Failed")), 2)

    def testE_FailJobs(self):
        """
        _FailJobs_

        Test our ability to fail jobs based on the information in the FWJR
        """
        workloadName = 'TestWorkload'

        self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', workloadName,
                                    'WMSandbox', 'WMWorkload.pkl')

        fwjrPath = os.path.join(WMCore.WMBase.getTestBase(),
                                "WMComponent_t/JobAccountant_t",
                                "fwjrs/badBackfillJobReport.pkl")

        testJobGroup = self.createTestJobGroup(nJobs=self.nJobs,
                                               workloadPath=workloadPath,
                                               fwjrPath=fwjrPath)

        badJobGroup = self.createTestJobGroup(nJobs=self.nJobs,
                                              workloadPath=workloadPath,
                                              fwjrPath=None,
                                              fileModifier='bad')

        config = self.getConfig()
        config.ErrorHandler.readFWJR = True
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')
        changer.propagate(badJobGroup.jobs, 'created', 'new')
        changer.propagate(badJobGroup.jobs, 'executing', 'created')
        changer.propagate(badJobGroup.jobs, 'complete', 'executing')
        changer.propagate(badJobGroup.jobs, 'jobfailed', 'complete')

        testErrorHandler = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler.reqAuxDB = None
        testErrorHandler.setup(None)
        testErrorHandler.exitCodesNoRetry = [8020]
        testErrorHandler.algorithm(None)

        # This should exhaust all jobs due to exit code
        # Except those with no fwjr
        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), 0)
        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), self.nJobs)
        idList = self.getJobs.execute(state='Exhausted')
        self.assertEqual(len(idList), self.nJobs)

        config.ErrorHandler.maxFailTime = -10
        testErrorHandler2 = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler2.reqAuxDB = None

        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')

        testErrorHandler2.algorithm(None)

        # This should exhaust all jobs due to timeout
        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), 0)
        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), self.nJobs)
        idList = self.getJobs.execute(state='Exhausted')
        self.assertEqual(len(idList), self.nJobs)

        config.ErrorHandler.maxFailTime = 24 * 3600
        config.ErrorHandler.passExitCodes = [8020]
        testErrorHandler3 = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler3.reqAuxDB = None

        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')

        testErrorHandler3.algorithm(None)

        # This should pass all jobs due to exit code
        idList = self.getJobs.execute(state='Created')
        self.assertEqual(len(idList), self.nJobs)

        return

    @attr('integration')
    def testZ_Profile(self):
        """
        _testProfile_

        Do a full profile of the poller
        """

        nJobs = 100
        workloadName = 'TestWorkload'
        self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', workloadName,
                                    'WMSandbox', 'WMWorkload.pkl')

        testJobGroup = self.createTestJobGroup(nJobs=nJobs, workloadPath=workloadPath)

        config = self.getConfig()
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'jobfailed', 'complete')

        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), nJobs)

        testErrorHandler = ErrorHandlerPoller(config)
        # set reqAuxDB None for the test,
        testErrorHandler.reqAuxDB = None
        testErrorHandler.setup(None)

        startTime = time.time()
        cProfile.runctx("testErrorHandler.algorithm()", globals(), locals(), filename="profStats.stat")
        stopTime = time.time()

        idList = self.getJobs.execute(state='CreateFailed')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='JobFailed')
        self.assertEqual(len(idList), 0)

        idList = self.getJobs.execute(state='JobCooloff')
        self.assertEqual(len(idList), nJobs)

        print("Took %f seconds to run polling algo" % (stopTime - startTime))

        p = pstats.Stats('profStats.stat')
        p.sort_stats('cumulative')
        p.print_stats(0.2)

        return


if __name__ == '__main__':
    unittest.main()



# python-future
from __future__ import print_function
from builtins import range

# system modules
import cherrypy
from multiprocessing import Process
from cheroot.test import webtest
from cherrypy import process
from threading import Thread, Condition
import time, random

# WMCore modules
from WMCore.REST.Test import setup_dummy_server, fake_authz_headers
from WMCore.REST.Test import fake_authz_key_file
from WMCore.REST.Server import RESTApi, RESTEntity, restcall

FAKE_FILE = fake_authz_key_file()
PORT = 8889

class Task(Thread):
    """A pseudo-task which runs in a separate thread. Provides standard
    functionality to start and stop with CherryPy engine transitions,
    and to get a "daemon state" object. In this case the status is just
    a series of time stamps when the daemon pretended to last do work."""

    def __init__(self):
        Thread.__init__(self, name = "Task")
        self._cv = Condition()
        self._status = {}
        self._stopme = False
        cherrypy.engine.subscribe("stop", self.stop)
        if cherrypy.engine.state == process.wspbus.states.STOPPED:
            cherrypy.engine.subscribe("start", self.start)
        else:
            self.start()

    def status(self):
        """Get the daemon status. Returns dictionary of time stamps of the
        the last hundred times this thread last did 'work'."""
        with self._cv:
            return dict((k, v) for k, v in self._status.items())

    def stop(self):
        """Tell the task thread to quit."""
        with self._cv:
            self._stopme = True
            self._cv.notifyAll()

    def run(self):
        """Run the task thread work routine. This just sleeps random [0,.5]
        seconds, and marks the time into the status dictionary, of course
        all in a thread-safe manner."""
        n = 0
        with self._cv:
            while not self._stopme:
                n += 1
                if n % 100 == 0:
                    self._status = {}
                    n = 0
                self._status[str(n)] = time.time()
                self._cv.wait(random.random() * 0.5)

class Status(RESTEntity):
    """REST entity to retrieve the status of running tasks."""
    def __init__(self, app, api, config, mount, tasks):
        """:arg list tasks: the list of task objects."""
        RESTEntity.__init__(self, app, api, config, mount)
        self._tasks = tasks

    def validate(self, *args):
        """No arguments, no validation required."""
        print("AMR validate executed")
        pass

    @staticmethod
    def gather(tasks):
        """Helper function which returns the status of each task,
        producing results as a generator."""
        for task in tasks:
            yield task.status()

    @restcall
    def get(self):
        """Get the status of all our registered tasks."""
        print("AMR status get request received")
        return self.gather(self._tasks)

class TaskAPI(RESTApi):
    """REST API which runs a bunch of pseudo tasks, and provides a REST
    entity to report their status via HTTP GET."""
    def __init__(self, app, config, mount):
        RESTApi.__init__(self, app, config, mount)
        print("AMR mounting RESTApi app: %s, config: %s, mount: %s" % (app, config, mount))
        tasks = [Task() for _ in range(0, 10)]
        self._add({ "status": Status(app, self, config, mount, tasks) })
        print("AMR done mounting the 'status' API")


class TaskTest(webtest.WebCase):
    """Client to verify :class:`TaskAPI` works."""
    def setUp(self):
        self.h = fake_authz_headers(FAKE_FILE.data)
        webtest.WebCase.PORT = PORT
        self.engine = cherrypy.engine
        self.proc = load_server(self.engine)
        print("AMR server loaded")

    def tearDown(self):
        stop_server(self.proc, self.engine)

    def test(self):
        h = self.h
        h.append(("Accept", "application/json"))
        print("AMR headers: %s" % h)
        print(self.getPage("/test", headers=h))
        print(self.getPage("/test/status", headers=h))
        for _ in range(0, 10):
            self.getPage("/test/status", headers=h)
            print(self.bodyY)
            time.sleep(.3)

def setup_server():
    srcfile = __file__.split("/")[-1].split(".py")[0]
    print("AMR srcfile : %s" % srcfile)
    setup_dummy_server(srcfile, "TaskAPI", authz_key_file=FAKE_FILE, port=PORT)

def load_server(engine):
    setup_server()
    proc = Process(target=start_server, name="cherrypy_Api_t", args=(engine,))
    proc.start()
    proc.join(timeout=1)
    return proc

def start_server(engine):
    webtest.WebCase.PORT = PORT
    cherrypy.log.screen = True
    engine.start()
    engine.block()

def stop_server(proc, engine):
    cherrypy.log.screen = True
    engine.stop()
    proc.terminate()

if __name__ == '__main__':
    webtest.main()
